# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import time
import scipy

from GridCal.Engine.Simulations.sparse_solve import get_sparse_type, get_linear_solver
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.ac_jacobian import AC_jacobian
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.common_functions import *
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import NumericPowerFlowResults
from GridCal.Engine.basic_structures import ReactivePowerControlMode
from GridCal.Engine.Simulations.PowerFlow.NumericalMethods.discrete_controls import control_q_inside_method

linear_solver = get_linear_solver()
sparse = get_sparse_type()
scipy.ALLOW_THREADS = True
np.set_printoptions(precision=8, suppress=True, linewidth=320)


def mu(Ybus, J, incS, dV, dx, pvpq, pq, npv, npq):
    """
    Calculate the Iwamoto acceleration parameter as described in:
    "A Load Flow Calculation Method for Ill-Conditioned Power Systems" by Iwamoto, S. and Tamura, Y."
    Args:
        Ybus: Admittance matrix
        J: Jacobian matrix
        incS: mismatch vector
        dV: voltage increment (in complex form)
        dx: solution vector as calculated dx = solve(J, incS)
        pvpq: array of the pq and pv indices
        pq: array of the pq indices

    Returns:
        the Iwamoto's optimal multiplier for ill conditioned systems
    """
    # evaluate the Jacobian of the voltage derivative
    # theoretically this is the second derivative matrix
    # since the Jacobian (J2) has been calculated with dV instead of V

    J2 = AC_jacobian(Ybus, dV, pvpq, pq, npv, npq)

    a = incS
    b = J * dx
    c = 0.5 * dx * J2 * dx

    g0 = -a.dot(b)
    g1 = b.dot(b) + 2 * a.dot(c)
    g2 = -3.0 * b.dot(c)
    g3 = 2.0 * c.dot(c)

    roots = np.roots([g3, g2, g1, g0])
    # three solutions are provided, the first two are complex, only the real solution is valid
    return roots[2].real


def IwamotoNR(Ybus, S0, V0, I0, Y0, pv_, pq_, Qmin, Qmax, tol, max_it=15,
              control_q=ReactivePowerControlMode.NoControl, robust=False) -> NumericPowerFlowResults:
    """
    Solves the power flow using a full Newton's method with the Iwamoto optimal step factor.
    Args:
        Ybus: Admittance matrix
        S0: Array of nodal power injections
        V0: Array of nodal voltages (initial solution)
        I0: Array of nodal current injections
        pv_: Array with the indices of the PV buses
        pq_: Array with the indices of the PQ buses
        tol: Tolerance
        max_it: Maximum number of iterations
        robust: use of the Iwamoto optimal step factor?.
    Returns:
        Voltage solution, converged?, error, calculated power injections

    @Author: Santiago Penate Vera
    """
    start = time.time()

    # initialize
    converged = 0
    iter_ = 0
    V = V0
    Va = np.angle(V)
    Vm = np.abs(V)
    dVa = np.zeros_like(Va)
    dVm = np.zeros_like(Vm)

    # set up indexing for updating V
    pq = pq_.copy()
    pv = pv_.copy()
    pvpq = np.r_[pv, pq]
    npv = len(pv)
    npq = len(pq)
    npvpq = npv + npq

    if npvpq > 0:

        # evaluate F(x0)
        Sbus = compute_zip_power(S0, I0, Y0, Vm)
        Scalc = compute_power(Ybus, V)
        f = compute_fx(Scalc, Sbus, pvpq, pq)
        norm_f = compute_fx_error(f)

        # check tolerance
        converged = norm_f < tol

        # do Newton iterations
        while not converged and iter_ < max_it:
            # update iteration counter
            iter_ += 1

            # evaluate Jacobian
            J = AC_jacobian(Ybus, V, pvpq, pq, npv, npq)

            # compute update step
            try:
                dx = linear_solver(J, f)
            except:
                print(J)
                converged = False
                iter_ = max_it + 1  # exit condition
                end = time.time()
                elapsed = end - start
                return NumericPowerFlowResults(V, converged, norm_f, Scalc,
                                               None, None, None, None, None, None, iter_, elapsed)

            # assign the solution vector
            dVa[pvpq] = dx[:npvpq]
            dVm[pq] = dx[npvpq:]
            dV = dVm * np.exp(1j * dVa)  # voltage mismatch

            # update voltage
            if robust:
                # if dV contains zeros will crash the second Jacobian derivative
                if not (dV == 0.0).any():
                    # calculate the optimal multiplier for enhanced convergence
                    mu_ = mu(Ybus, J, f, dV, dx, pvpq, pq, npv, npq)
                else:
                    mu_ = 1.0
            else:
                mu_ = 1.0

            Vm -= mu_ * dVm
            Va -= mu_ * dVa
            V = Vm * np.exp(1j * Va)

            Vm = np.abs(V)  # update Vm and Va again in case
            Va = np.angle(V)  # we wrapped around with a negative Vm

            # evaluate F(x)
            Sbus = compute_zip_power(S0, I0, Y0, Vm)
            Scalc = compute_power(Ybus, V)
            f = compute_fx(Scalc, Sbus, pvpq, pq)
            norm_f = compute_fx_error(f)

            # review reactive power limits
            # it is only worth checking Q limits with a low error
            # since with higher errors, the Q values may be far from realistic
            # finally, the Q control only makes sense if there are pv nodes
            if control_q != ReactivePowerControlMode.NoControl and norm_f < 1e-2 and npv > 0:

                # check and adjust the reactive power
                # this function passes pv buses to pq when the limits are violated,
                # but not pq to pv because that is unstable
                n_changes, Scalc, Sbus, pv, pq, pvpq, messages = control_q_inside_method(Scalc, Sbus, pv, pq,
                                                                                         pvpq, Qmin, Qmax)

                if n_changes > 0:
                    # adjust internal variables to the new pq|pv values
                    npv = len(pv)
                    npq = len(pq)
                    npvpq = npv + npq

                    # recompute the error based on the new Sbus
                    f = compute_fx(Scalc, Sbus, pvpq, pq)
                    norm_f = compute_fx_error(f)

                    # if verbose > 0:
                    #     for sense, idx, var in messages:
                    #         msg = "Bus " + str(idx) + " changed to PQ, limited to " + str(var * 100) + " MVAr"
                    #         logger.add_debug(msg)

            # check convergence
            converged = norm_f < tol

            # check for absurd values
            if np.isnan(V).any() or (Vm == 0).any():
                break

    else:
        norm_f = 0
        converged = True
        Scalc = S0 + I0 * Vm + Y0 * np.power(Vm, 2)  # compute the ZIP power injection

    end = time.time()
    elapsed = end - start

    return NumericPowerFlowResults(V, converged, norm_f, Scalc, None, None, None, None, None, None, iter_, elapsed)




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os.path

import numpy as np

from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.basic_structures import BranchImpedanceMode
from GridCal.Engine.basic_structures import BusMode
from GridCal.Engine.Devices.enumerations import ConverterControlType, TransformerControlType
from GridCal.Engine.Devices import *
from GridCal.Engine.basic_structures import Logger, SolverType, ReactivePowerControlMode, TapsControlMode
from GridCal.Engine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions
from GridCal.Engine.Simulations.PowerFlow.power_flow_results import PowerFlowResults


try:
    import power_grid_model as pgm
    from power_grid_model.validation import validate_input_data, assert_valid_input_data, ValidationError, ValidationException

    ALLIANDER_PGM_AVAILABLE = True
    print("Alliander's PGM available")

except ImportError:
    ALLIANDER_PGM_AVAILABLE = False
    print("Alliander's power grid model is not available, try pip install power-grid-model")

'''
hierarchy

base ──┬─────────────────────────────────────────────── node
       │
       ├── branch ──────────────────────────────────┬── line
       │                                            ├── link
       │                                            └── transformer
       │
       ├── appliance ──┬─────────────────────────────── source
       │               │
       │               ├─────────────────────────────── shunt
       │               │
       │               └── generic_load_gen ────────┬── sym_load
       │                                            ├── sym_gen
       │                                            ├── asym_load
       │                                            └── asym_gen
       │
       └── sensor ─────┬── generic_voltage_sensor ──┬── sym_voltage_sensor
                       │                            └── asym_voltage_sensor
                       │
                       └── generic_power_sensor ────┬── sym_power_sensor
                                                    └── asym_power_sensor

'''



def get_pgm_buses(circuit: MultiCircuit, idx0):
    """
    Convert the buses to Alliander's PGM buses
    :param circuit: GridCal circuit
    :return: bus dictionary buses[uuid] -> int
    """
    bus_dict = dict()

    node = pgm.initialize_array('input', 'node', len(circuit.buses))
    idx = idx0
    for i, bus in enumerate(circuit.buses):

        # fill in data
        node['id'][i] = idx
        node['u_rated'][i] = bus.Vnom * 1000.0  # in V

        # create dictionary entry
        bus_dict[bus.idtag] = idx
        idx += 1

    return node, bus_dict, idx


def get_pgm_loads(circuit: MultiCircuit, bus_dict, idx0):
    """
    Generate load data
    :param circuit: GridCal circuit
    :param bus_dict: dictionary of bus id to Alliander's PGM index
    :return struct of load values
    """

    devices = circuit.get_loads()

    sym_load = pgm.initialize_array('input', 'sym_load', len(devices) * 3)

    idx = idx0
    for k, elm in enumerate(devices):
        k1 = 3 * k
        k2 = 3 * k + 1
        k3 = 3 * k + 2

        sym_load['id'][k1] = idx
        sym_load['node'][k1] = bus_dict[elm.bus.idtag]
        sym_load['status'][k1] = int(elm.active)
        sym_load['type'][k1] = pgm.LoadGenType.const_power
        sym_load['p_specified'][k1] = elm.P * 1e6
        sym_load['q_specified'][k1] = elm.Q * 1e6
        idx += 1

        sym_load['id'][k2] = idx
        sym_load['node'][k2] = bus_dict[elm.bus.idtag]
        sym_load['status'][k2] = int(elm.active)
        sym_load['type'][k2] = pgm.LoadGenType.const_current
        sym_load['p_specified'][k2] = elm.Ir * 1e6
        sym_load['q_specified'][k2] = elm.Ii * 1e6
        idx += 1

        sym_load['id'][k3] = idx
        sym_load['node'][k3] = bus_dict[elm.bus.idtag]
        sym_load['status'][k3] = int(elm.active)
        sym_load['type'][k3] = pgm.LoadGenType.const_impedance
        sym_load['p_specified'][k3] = elm.G * 1e6
        sym_load['q_specified'][k3] = elm.B * 1e6
        idx += 1

    return sym_load, idx


# def get_pgm_static_generators(circuit: MultiCircuit, bus_dict, idx0):
#     """
#
#     :param circuit: GridCal circuit
#     :param bus_dict: dictionary of bus id to Alliander's PGM index
#     """
#     devices = circuit.get_static_generators()
#
#     stagen = pgm.initialize_array('input', 'sym_gen', len(devices))
#
#     idx = idx0
#     for k, elm in enumerate(devices):
#         stagen['id'][k] = idx
#         stagen['node'][k] = bus_dict[elm.bus.idtag]
#         stagen['status'][k] = int(elm.active)
#         stagen['type'][k] = pgm.LoadGenType.const_power
#         stagen['p_specified'][k] = elm.P * 1e6
#         stagen['q_specified'][k] = elm.Q * 1e6
#         idx += 1
#
#     return stagen, idx


def get_pgm_shunts(circuit: MultiCircuit, bus_dict, idx0):
    """

    :param circuit: GridCal circuit
    :param bus_dict: dictionary of bus id to Alliander's PGM index
    """
    devices = circuit.get_shunts()

    shunt = pgm.initialize_array('input', 'shunt', len(devices))

    idx = idx0
    for k, elm in enumerate(devices):
        Ybase = circuit.Sbase / (elm.bus.Vnom**2)

        shunt['id'][k] = idx
        shunt['node'][k] = bus_dict[elm.bus.idtag]
        shunt['status'][k] = int(elm.active)

        shunt['g1'][k] = elm.G * Ybase
        shunt['b1'][k] = elm.B * Ybase
        idx += 1

    return shunt, idx


def get_pgm_generators(circuit: MultiCircuit, bus_dict, idx0):
    """

    :param circuit: GridCal circuit
    :param bus_dict: dictionary of bus id to Alliander's PGM index
    """
    gen_devices = circuit.get_generators()
    stagen_devices = circuit.get_static_generators()
    batt_devices = circuit.get_batteries()

    sym_gen = pgm.initialize_array('input', 'sym_gen',
                                   len(gen_devices) + len(stagen_devices) + len(batt_devices))

    idx = idx0

    for k, elm in enumerate(gen_devices + batt_devices):
        sym_gen['id'][k] = idx
        sym_gen['node'][k] = bus_dict[elm.bus.idtag]
        sym_gen['status'][k] = int(elm.active)
        sym_gen['type'][k] = pgm.LoadGenType.const_power
        sym_gen['p_specified'][k] = elm.P * 1e6
        sym_gen['q_specified'][k] = 0
        idx += 1

    k2 = len(gen_devices) + len(batt_devices)
    for k, elm in enumerate(stagen_devices):
        sym_gen['id'][k2] = idx
        sym_gen['node'][k2] = bus_dict[elm.bus.idtag]
        sym_gen['status'][k2] = int(elm.active)
        sym_gen['type'][k2] = pgm.LoadGenType.const_power
        sym_gen['p_specified'][k2] = elm.P * 1e6
        sym_gen['q_specified'][k2] = elm.Q * 1e6
        idx += 1
        k2 += 1

    return sym_gen, idx


def get_pgm_source(circuit: MultiCircuit, bus_dict, idx0):
    """

    :param circuit: GridCal circuit
    :param bus_dict: dictionary of bus id to Alliander's PGM index
    """
    gen_devices = circuit.get_generators()

    sym_gen = pgm.initialize_array('input', 'source', 1)

    # TODO: pick the one with the largest P
    elm = gen_devices[0]

    idx = idx0
    k = 0
    sym_gen['id'][k] = idx
    sym_gen['node'][k] = bus_dict[elm.bus.idtag]
    sym_gen['status'][k] = int(elm.active)
    sym_gen['u_ref'][k] = elm.Vset
    # sym_gen['u_ref_angle'][k] = 0.0
    idx += 1

    return sym_gen, idx


# def get_pgm_battery_data(circuit: MultiCircuit, bus_dict):
#     """
#
#     :param circuit: GridCal circuit
#     :param bus_dict: dictionary of bus id to Alliander's PGM index
#     """
#     devices = circuit.get_batteries()
#     batt = pgm.initialize_array('input', 'sym_load', len(devices))
#     for k, elm in enumerate(devices):
#         pass
#     return batt


def get_pgm_line(circuit: MultiCircuit, bus_dict, idx0, logger: Logger):
    """

    :param circuit: GridCal circuit
    :param bus_dict: dictionary of bus id to Alliander's PGM index
    """

    line = pgm.initialize_array('input', 'line', len(circuit.lines))
    omega = 6.283185307 * circuit.fBase  # angular frequency
    r3 = np.sqrt(3.0)  # square root of 3

    # Compile the lines
    idx = idx0
    for i, elm in enumerate(circuit.lines):
        Vf = elm.bus_from.Vnom
        Vt = elm.bus_to.Vnom

        if Vf != Vt:
            logger.add_error('Different line terminal voltages', elm.name, str(Vt), str(Vf))
            elm.bus_to.Vnom = Vf

        Zbase = Vf * Vf / circuit.Sbase

        line['id'][i] = idx
        line['from_node'][i] = bus_dict[elm.bus_from.idtag]
        line['to_node'][i] = bus_dict[elm.bus_to.idtag]
        line['from_status'][i] = int(elm.bus_from.active)
        line['to_status'][i] = int(elm.bus_to.active)
        line['r1'][i] = elm.R * Zbase  # Ohm
        line['x1'][i] = elm.X * Zbase  # Ohm
        line['c1'][i] = elm.B / (omega * Zbase)  # Farad
        line['tan1'][i] = 0.0  # this is the ratio G/B, which does not apply here because we do not have G
        line['i_n'][i] = 1e6 * elm.rate / r3 / (Vf * 1000)  # rating in A
        idx += 1

    return line, idx


def get_pgm_transformer_data(circuit: MultiCircuit, bus_dict, idx0):
    """

    :param circuit: GridCal circuit
    :param bus_dict: dictionary of bus id to Alliander's PGM index
    """
    xfo = pgm.initialize_array('input', 'transformer', len(circuit.transformers2w))

    omega = 6.283185307 * circuit.fBase
    r3 = np.sqrt(3.0)

    # Compile the lines
    idx = idx0
    for i, elm in enumerate(circuit.transformers2w):
        Zbase = elm.bus_from.Vnom ** 2 / circuit.Sbase

        xfo['id'][i] = idx
        xfo['from_node'][i] = bus_dict[elm.bus_from.idtag]
        xfo['to_node'][i] = bus_dict[elm.bus_to.idtag]
        xfo['from_status'][i] = int(elm.bus_from.active)
        xfo['to_status'][i] = int(elm.bus_to.active)

        xfo['u1'][i] = 1e3 * elm.bus_from.Vnom  # rated voltage at from-side (V)
        xfo['u2'][i] = 1e3 * elm.bus_to.Vnom  # rated voltage at to-side (V)
        xfo['sn'][i] = 1e6 * elm.rate  # volt-ampere (VA)
        xfo['uk'][i] = elm.X * (elm.rate / circuit.Sbase)  # relative short circuit voltage (p.u.)
        xfo['pk'][i] = 0  # short circuit (copper) loss (W)
        xfo['i0'][i] = 0  # relative no-load current (p.u.)
        xfo['p0'][i] = 0  # no-load (iron) loss (W)


        # clock number of phase shift.
        # Even number is not possible if one side is Y(N)
        # winding and the other side is not Y(N) winding.
        # Odd number is not possible, if both sides are Y(N)
        # winding or both sides are not Y(N) winding.
        phase_int = int(np.round(np.rad2deg(elm.angle) / 30)) % 12
        xfo['clock'][i] = phase_int

        xfo['winding_from'][i] = pgm.WindingType.wye  # WindingType object

        if phase_int % 2 == 0:
            xfo['winding_to'][i] = pgm.WindingType.wye_n  # WindingType object
        else:
            xfo['winding_to'][i] = pgm.WindingType.delta  # WindingType object

        xfo['tap_side'][i] = pgm.BranchSide.to_side  # BranchSide object
        xfo['tap_pos'][i] = 0  # current position of tap changer
        xfo['tap_min'][i] = 0  # position of tap changer at minimum voltage
        xfo['tap_max'][i] = 0  # position of tap changer at maximum voltage
        xfo['tap_nom'][i] = 0  # nominal position of tap changer
        xfo['tap_size'][i] = 1e-20  # size of each tap of the tap changer (V), TODO: this should be 0 in future versions

        # xfo['uk_min'][i] = 0  # relative short circuit voltage at minimum tap
        # xfo['uk_max'][i] = 0  # relative short circuit voltage at maximum tap
        # xfo['pk_min'][i] = 0  # short circuit (copper) loss at minimum tap (W)
        # xfo['pk_max'][i] = 0  # short circuit (copper) loss at maximum tap (W)
        #
        # xfo['r_grounding_from'][i] = 0  # grounding resistance at from-side, if relevant (Ohm)
        # xfo['x_grounding_from'][i] = 0  # grounding reactance at from-side, if relevant (Ohm)
        # xfo['r_grounding_to'][i] = 0  # grounding resistance at to-side, if relevant (Ohm)
        # xfo['x_grounding_to'][i] = 0  # grounding reactance at to-side, if relevant (Ohm)
        idx += 1

    return xfo, idx


def get_pgm_vsc_data(circuit: MultiCircuit, bus_dict, idx0):
    """

    :param circuit: GridCal circuit
    :param bus_dict: dictionary of bus id to Alliander's PGM index
    """
    vsc = pgm.initialize_array('input', 'sym_load', len(circuit.vsc_devices))
    idx = idx0
    for i, elm in enumerate(circuit.vsc_devices):
        pass
    return vsc, idx


def get_pgm_dc_line_data(circuit: MultiCircuit, bus_dict, idx0):
    """

    :param circuit: GridCal circuit
    :param bus_dict: dictionary of bus id to Alliander's PGM index
    """
    dc_line = pgm.initialize_array('input', 'sym_load', len(circuit.dc_lines))
    # Compile the lines
    idx = idx0
    for i, elm in enumerate(circuit.dc_lines):
        pass
    return dc_line, idx


def get_pgm_hvdc_data(circuit: MultiCircuit, bus_dict, idx0):
    """

    :param circuit: GridCal circuit
    :param bus_dict: dictionary of bus id to Alliander's PGM index
    """
    hvdc = pgm.initialize_array('input', 'sym_load', len(circuit.hvdc_lines))
    idx = idx0
    for i, elm in enumerate(circuit.hvdc_lines):
        pass
    return hvdc, idx


def to_pgm(circuit: MultiCircuit, logger: Logger = Logger()) -> "pgm.PowerGridModel":
    """
    Convert GridCal circuit to Alliander's PGM model
    See https://github.com/alliander-opensource/power-grid-model/blob/main/docs/graph-data-model.md
    :param circuit: MultiCircuit
    :return: pgm.PowerGridModel instance
    """
    idx0 = 0
    node, bus_dict, idx0 = get_pgm_buses(circuit, idx0)

    sym_load, idx0 = get_pgm_loads(circuit, bus_dict, idx0)

    shunt, idx0 = get_pgm_shunts(circuit, bus_dict, idx0)
    sym_gen, idx0 = get_pgm_generators(circuit, bus_dict, idx0)
    source, idx0 = get_pgm_source(circuit, bus_dict, idx0)
    line, idx0 = get_pgm_line(circuit, bus_dict, idx0, logger)
    transformer, idx0 = get_pgm_transformer_data(circuit, bus_dict, idx0)

    # vsc = get_pgm_vsc_data(circuit, bus_dict)
    # dc_line = get_pgm_dc_line_data(circuit, bus_dict)
    # hvdc = get_pgm_hvdc_data(circuit, bus_dict)

    # all
    input_data = {
        'node': node,
        'line': line,
        'transformer': transformer,
        'sym_load': sym_load,
        'sym_gen': sym_gen,
        'source': source,
        'shunt': shunt
    }

    try:
        assert_valid_input_data(input_data=input_data)
        model_ok = True
    except ValidationException as e:
        print(e)
        model_ok = False

    if model_ok:
        model = pgm.PowerGridModel(input_data, system_frequency=circuit.fBase)
    else:
        model = None

    return model


def alliander_pgm_pf(circuit: MultiCircuit, opt: PowerFlowOptions, logger: Logger):
    """
    Alliander's PGM power flow
    :param circuit: MultiCircuit instance
    :param opt: Power Flow Options
    :param logger: Logger object
    :return: Alliander's PGM Power flow results object
    """
    model = to_pgm(circuit, logger=logger)

    try:
        pf_res = model.calculate_power_flow()

    except RuntimeError as e:
        logger.add_error('Power flow failed\n' + str(e))
        pf_res = None

    gc_res = translate_pgm_results(circuit, pf_res)

    return gc_res


def translate_pgm_results(grid: MultiCircuit, pf_res) -> PowerFlowResults:
    from GridCal.Engine.Core.snapshot_pf_data import compile_snapshot_circuit

    nc = compile_snapshot_circuit(grid)

    results = PowerFlowResults(n=nc.nbus,
                               m=nc.nbr,
                               n_tr=nc.ntr,
                               n_hvdc=nc.nhvdc,
                               bus_names=nc.bus_names,
                               branch_names=nc.branch_names,
                               transformer_names=nc.transformer_data.tr_names,
                               hvdc_names=nc.hvdc_names,
                               bus_types=nc.bus_types)

    if pf_res is None:
        return results

    Vm = pf_res['node']['u_pu']
    Va = np.zeros_like(Vm)
    Pf = pf_res['line']['p_from'] * 1e-6
    Pt = pf_res['line']['p_to'] * 1e-6
    Qf = pf_res['line']['q_from'] * 1e-6
    Qt = pf_res['line']['q_to'] * 1e-6

    if 'transformer' in pf_res:
        Pf = np.r_[Pf, pf_res['transformer']['p_from']] * 1e-6
        Pt = np.r_[Pt, pf_res['transformer']['p_to']] * 1e-6
        Qf = np.r_[Pf, pf_res['transformer']['q_from']] * 1e-6
        Qt = np.r_[Pt, pf_res['transformer']['q_to']] * 1e-6

    losses = (Pf + Pt) + 1j * (Qf + Qt)

    results.voltage = Vm * np.exp(1j * Va)
    # results.Sbus = res.S[0, :]
    results.Sf = Pf + 1j * Qf
    results.St = Pt + 1j * Qt
    results.loading = Pf / (nc.branch_rates + 1e-20)
    results.losses = losses
    # results.Vbranch = res.Vbranch[0, :]
    # results.If = res.If[0, :]
    # results.It = res.It[0, :]
    # results.Beq = res.Beq[0, :]
    # results.m = res.tap_modules[0, :]
    # results.theta = res.tap_angles[0, :]

    results.F = nc.F
    results.T = nc.T
    # results.hvdc_F = res.F_hvdc
    # results.hvdc_T = res.T_hvdc
    # results.hvdc_Pf = res.hvdc_Pf[0, :]
    # results.hvdc_Pt = res.hvdc_Pt[0, :]
    # results.hvdc_loading = res.hvdc_loading[0, :]
    # results.hvdc_losses = res.hvdc_losses[0, :]
    results.bus_area_indices = grid.get_bus_area_indices()
    results.area_names = [a.name for a in grid.areas]

    return results


if __name__ == "__main__":
    import GridCal.Engine as gce
    # fname = './../../../../../Grids_and_profiles/grids/IEEE 14.xlsx'
    fname = './../../../../../Grids_and_profiles/grids/IEEE 30 Bus.gridcal'
    circ = gce.FileOpen(fname).open()

    pf_opt = PowerFlowOptions()
    lgr = Logger()
    pgm_ = alliander_pgm_pf(circ, pf_opt, lgr)
    print()



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from io import StringIO, TextIOWrapper, BytesIO
import os
import numpy as np
import chardet
import pandas as pd
import zipfile
from typing import List, Dict, Any
from GridCal.Engine.basic_structures import Logger
from GridCal.Engine.IO.generic_io_functions import parse_config_df
# from GridCal.Gui.Session.session import SimulationSession


def save_data_frames_to_zip(dfs: Dict[str, pd.DataFrame], filename_zip="file.zip",
                            text_func=None, progress_func=None,
                            sessions: List[Any] = []):
    """
    Save a list of DataFrames to a zip file without saving to disk the csv files
    :param dfs: dictionary of pandas dataFrames {name: DataFrame}
    :param filename_zip: file name where to save all
    :param text_func: pointer to function that prints the names
    :param progress_func: pointer to function that prints the progress 0~100
    :param sessions: SimulationSession instance
    """

    n = len(dfs)
    n_failed = 0
    # open zip file for writing
    with zipfile.ZipFile(filename_zip, 'w', zipfile.ZIP_DEFLATED) as f_zip_ptr:

        # for each DataFrame and name...
        i = 0
        for name, df in dfs.items():

            if text_func is not None:
                text_func('Flushing ' + name + ' to ' + filename_zip + '...')

            if progress_func is not None:
                progress_func((i + 1) / n * 100)

            if name.endswith('_prof'):

                # compose the csv file name
                filename = name + ".parquet"

                # open a string buffer
                try:  # try parquet file
                    with BytesIO() as buffer:
                        # save the DataFrame to the buffer, protocol4 is to be compatible with python 3.6
                        df.to_parquet(buffer)
                        # save the buffer to the zip file
                        f_zip_ptr.writestr(filename, buffer.getvalue())

                except:  # otherwise just use csv
                    n_failed += 1
                    filename = name + ".csv"
                    with StringIO() as buffer:
                        df.to_csv(buffer, index=False)  # save the DataFrame to the buffer
                        f_zip_ptr.writestr(filename, buffer.getvalue())  # save the buffer to the zip file
            else:
                # compose the csv file name
                filename = name + ".csv"

                # open a string buffer
                with StringIO() as buffer:
                    df.to_csv(buffer, index=False)  # save the DataFrame to the buffer
                    f_zip_ptr.writestr(filename, buffer.getvalue())  # save the buffer to the zip file

            i += 1

        # pre-count the sessions
        n_items = 0
        for session in sessions:
            for drv_name, drv in session.drivers.items():
                if hasattr(drv, 'results'):
                    if drv.results is not None:
                        for arr_name, arr in drv.results.get_arrays().items():
                            n_items += 1

        # save sessions
        i = 0
        for session in sessions:
            for drv_name, drv in session.drivers.items():
                if hasattr(drv, 'results'):
                    if drv.results is not None:
                        for arr_name, arr in drv.results.get_arrays().items():
                            filename = 'sessions/' + session.name + '/' + drv.tpe.value + '/' + arr_name

                            if text_func is not None:
                                text_func('Flushing ' + filename + ' to ' + filename_zip + '...')

                            with BytesIO() as buffer:
                                # np.save(buffer, np.array(arr))  # save the DataFrame to the buffer
                                # f_zip_ptr.writestr(filename + '.npy', buffer.getvalue())  # save the buffer to the zip file

                                # save the DataFrame to the buffer, protocol4 is to be compatible with python 3.6
                                np.save(buffer, np.array(arr))  # save the DataFrame to the buffer
                                f_zip_ptr.writestr(filename + '.npy', buffer.getvalue())  # save the buffer to the zip file

                            if progress_func is not None:
                                progress_func((i + 1) / n_items * 100)

                            i += 1

    if n_failed:
        print('Failed to pickle several profiles, but saved them as csv.\nFor improved speed install Pandas >= 1.2')


def read_data_frame_from_zip(file_pointer, extension, index_col=None, logger=Logger()):
    """
    read DataFrame
    :param file_pointer: Pointer to the file within the zip file
    :param extension: Extension, just to determine the reader method
    :param index_col: Index col (only for config file)
    :param logger:
    :return: DataFrame
    """
    try:
        if extension == '.csv':
            return pd.read_csv(file_pointer, index_col=index_col)

        elif extension == '.npy':
            try:
                return np.load(file_pointer)
            except ValueError:
                return np.load(file_pointer, allow_pickle=True)

        elif extension == '.pkl':
            try:
                return pd.read_pickle(file_pointer)
            except ValueError as e:
                logger.add_error(str(e), device=file_pointer.name)
                return None
            except AttributeError as e:
                logger.add_error(str(e) + ' Upgrading pandas might help.', device=file_pointer.name)
                return None

        elif extension == '.parquet':
            try:
                return pd.read_parquet(file_pointer)
            except ValueError as e:
                logger.add_error(str(e), device=file_pointer.name)
                return None
            except AttributeError as e:
                logger.add_error(str(e) + ' Upgrading pandas might help.', device=file_pointer.name)
                return None

    except EOFError:
        return None
    except zipfile.BadZipFile:
        return None


def get_frames_from_zip(file_name_zip, text_func=None, progress_func=None, logger=Logger()):
    """
    Open the csv files from a zip file
    :param file_name_zip: name of the zip file
    :param text_func: pointer to function that prints the names
    :param progress_func: pointer to function that prints the progress 0~100
    :param logger:
    :return: list of DataFrames
    """

    # open the zip file
    try:
        zip_file_pointer = zipfile.ZipFile(file_name_zip)
    except zipfile.BadZipFile:
        return None

    names = zip_file_pointer.namelist()

    n = len(names)
    data = dict()

    # for each file in the zip file...
    for i, file_name in enumerate(names):

        # split the file name into name and extension
        name, extension = os.path.splitext(file_name)

        if text_func is not None:
            text_func('Unpacking ' + name + ' from ' + file_name_zip)

        if progress_func is not None:
            progress_func((i + 1) / n * 100)

        # create a buffer to read the file
        file_pointer = zip_file_pointer.open(file_name)

        if name.lower() == "config":
            df = read_data_frame_from_zip(file_pointer, extension, index_col=0, logger=logger)
            data = parse_config_df(df, data)
        else:
            # make pandas read the file
            df = read_data_frame_from_zip(file_pointer, extension, logger=logger)

        # append the DataFrame to the list
        if df is not None:
            data[name] = df

    return data


def get_session_tree(file_name_zip: str):
    """
    Get the sessions structure
    :param file_name_zip:
    :return:
    """
    try:
        zip_file_pointer = zipfile.ZipFile(file_name_zip)
    except zipfile.BadZipFile:
        return dict()

    names = zip_file_pointer.namelist()

    data = dict()

    for name in names:
        if '/' in name:
            path = name.split('/')
            if path[0].lower() == 'sessions':

                session_name = path[1]
                study_name = path[2]
                array_name = path[3]

                if session_name not in data.keys():
                    data[session_name] = dict()

                if study_name not in data[session_name].keys():
                    data[session_name][study_name] = list()

                data[session_name][study_name].append(array_name)

    return data


def load_session_driver_objects(file_name_zip: str, session_name: str, study_name: str):
    """
    Get the sessions structure
    :param file_name_zip:
    :param session_name:
    :param study_name:
    :return:
    """
    try:
        zip_file_pointer = zipfile.ZipFile(file_name_zip)
    except zipfile.BadZipFile:
        return dict()

    data = dict()

    # traverse the zip names and pick all those that start with sessions/session_name/study_name
    for name in zip_file_pointer.namelist():
        if '/' in name:
            path = name.split('/')
            if len(path) > 3:
                if path[0].lower() == 'sessions' and session_name == path[1] and study_name == path[2]:

                    # create a buffer to read the file
                    file_pointer = zip_file_pointer.open(name)

                    # split the file name into name and extension
                    _, extension = os.path.splitext(name)
                    arr_name = path[3].replace(extension, '')

                    # read the data
                    data[arr_name] = read_data_frame_from_zip(file_pointer, extension)

                    # try:
                    #     data[arr_name] = np.load(file_pointer)
                    # except ValueError:
                    #     data[arr_name] = np.load(file_pointer, allow_pickle=True)

    return data


def get_xml_content(file_ptr):
    """
    Reads the content of a file
    :param file_ptr: File pointer (from file or zip file)
    :return: list of text lines
    """
    # xml files always have the encoding declared, find it out
    first_line = file_ptr.readline()

    if b'encoding' in first_line:
        encoding = first_line.split()[2].split(b'=')[1].replace(b'"', b'').replace(b'?>', b'').decode()
    else:
        try:
            detection = chardet.detect(first_line)
            encoding = detection['encoding']
        except:
            encoding = 'utf-8'

    # sequential back to the start
    file_ptr.seek(0)

    # read all the lines
    with TextIOWrapper(file_ptr, encoding=encoding) as fle:
        text_lines = [l for l in fle]

    return text_lines


def get_xml_from_zip(file_name_zip, text_func=None, progress_func=None):
    """
    Get the .xml files from a zip file
    :param file_name_zip: name of the zip file
    :param text_func: pointer to function that prints the names
    :param progress_func: pointer to function that prints the progress 0~100
    :return: list of xml file contents
    """

    # open the zip file
    try:
        zip_file_pointer = zipfile.ZipFile(file_name_zip)
    except zipfile.BadZipFile:
        return None

    names = zip_file_pointer.namelist()

    n = len(names)
    data = dict()

    # for each file in the zip file...
    for i, file_name in enumerate(names):

        # split the file name into name and extension
        name, extension = os.path.splitext(file_name)

        if text_func is not None:
            text_func('Unpacking ' + name + ' from ' + file_name_zip)

        if progress_func is not None:
            progress_func((i + 1) / n * 100)

        if extension == '.xml':
            file_ptr = zip_file_pointer.open(file_name)

            text_lines = get_xml_content(file_ptr)

            data[name] = text_lines

    return data


if __name__ == '__main__':

    fname = '/home/santi/Descargas/IEEE 39.gridcal'
    # data = get_xml_from_zip(fname)

    data_ = get_session_tree(fname)

    load_session_driver_objects(file_name_zip=fname,
                                session_name='GUI session',
                                study_name='Power Flow')

    print()


# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import time
import datetime
import numpy as np
import pandas as pd
from numba import jit, prange
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Core.time_series_pf_data import compile_time_circuit
from GridCal.Engine.Simulations.LinearFactors.linear_analysis import LinearAnalysis
from GridCal.Engine.Simulations.ContingencyAnalysis.contingency_analysis_driver import ContingencyAnalysisOptions
from GridCal.Engine.Simulations.ContingencyAnalysis.contingency_analysis_ts_results import ContingencyAnalysisTimeSeriesResults
from GridCal.Engine.Simulations.driver_types import SimulationTypes
from GridCal.Engine.Simulations.driver_template import TimeSeriesDriverTemplate


@jit(nopython=True, parallel=False, cache=True)
def compute_flows_numba_t(e, c, nt, LODF, Flows, rates, overload_count, max_overload, worst_flows):
    """
    Compute LODF based flows (Sf)
    :param e: element index
    :param c: contingency element index
    :param nt: number of time steps
    :param LODF: LODF matrix (element, failed element)
    :param Flows: base Sf matrix (time, element)
    :param rates: Matrix of rates (time, element)
    :param overload_count: [out] number of overloads per element (element, contingency element)
    :param max_overload: [out] maximum overload per element (element, contingency element)
    :param worst_flows: [out] worst flows per element (time, element)
    :return: Cube of N-1 Flows (time, elements, contingencies)
    """

    for t in range(nt):
        # the formula is: Fn-1(i) = Fbase(i) + LODF(i,j) * Fbase(j) here i->line, j->failed line
        flow_n_1 = LODF[e, c] * Flows[t, c] + Flows[t, e]
        flow_n_1_abs = abs(flow_n_1)

        if rates[t, e] > 0:
            rate = flow_n_1_abs / rates[t, e]

            if rate > 1:
                overload_count[e, c] += 1
                if flow_n_1_abs > max_overload[e, c]:
                    max_overload[e, c] = flow_n_1_abs

        if flow_n_1_abs > abs(worst_flows[t, e]):
            worst_flows[t, e] = flow_n_1


@jit(nopython=True, parallel=True, cache=True)
def compute_flows_numba(e, nt, contingency_branch_idx, LODF, Flows, rates, overload_count,
                        max_overload, worst_flows, parallelize_from=500):
    """
    Compute LODF based Sf
    :param e: element index
    :param nt: number of time steps
    :param contingency_branch_idx: list of branch indices to fail
    :param LODF: LODF matrix (element, failed element)
    :param Flows: base Sf matrix (time, element)
    :param rates:
    :param overload_count:
    :param max_overload:
    :param worst_flows:
    :param parallelize_from:
    :return:  Cube of N-1 Flows (time, elements, contingencies)
    """
    nc = len(contingency_branch_idx)
    if nc < parallelize_from:
        for ic in range(nc):
            c = contingency_branch_idx[ic]
            compute_flows_numba_t(e, c, nt, LODF, Flows, rates, overload_count, max_overload, worst_flows)
    else:
        for ic in prange(nc):
            c = contingency_branch_idx[ic]
            compute_flows_numba_t(e, c, nt, LODF, Flows, rates, overload_count, max_overload, worst_flows)


class ContingencyAnalysisTimeSeries(TimeSeriesDriverTemplate):
    name = 'Contingency analysis time series'
    tpe = SimulationTypes.ContingencyAnalysisTS_run

    def __init__(self, grid: MultiCircuit, options: ContingencyAnalysisOptions, start_=0, end_=None):
        """
        N - k class constructor
        @param grid: MultiCircuit Object
        @param options: N-k options
        @:param pf_options: power flow options
        """
        TimeSeriesDriverTemplate.__init__(self,
                                          grid=grid,
                                          start_=start_,
                                          end_=end_)

        # Options to use
        self.options = options

        # N-K results
        self.results = ContingencyAnalysisTimeSeriesResults(n=0, ne=0, nc=0,
                                                            time_array=(),
                                                            bus_names=(),
                                                            branch_names=(),
                                                            bus_types=())

        self.branch_names = list()

    def get_steps(self):
        """
        Get variations list of strings
        """
        if self.results is not None:
            return [l.strftime('%d-%m-%Y %H:%M') for l in
                    pd.to_datetime(self.grid.time_profile[self.start_: self.end_])]
        else:
            return list()

    def n_minus_k(self):
        """
        Run N-1 simulation in series
        :return: returns the results
        """

        self.progress_text.emit("Analyzing...")

        ts_numeric_circuit = compile_time_circuit(self.grid)
        ne = ts_numeric_circuit.nbr
        nc = ts_numeric_circuit.nbr
        nt = len(ts_numeric_circuit.time_array)

        results = ContingencyAnalysisTimeSeriesResults(ne=ne,
                                                       nc=nc,
                                                       time_array=ts_numeric_circuit.time_array,
                                                       n=ts_numeric_circuit.nbus,
                                                       branch_names=ts_numeric_circuit.branch_names,
                                                       bus_names=ts_numeric_circuit.bus_names,
                                                       bus_types=ts_numeric_circuit.bus_types)

        linear_analysis = LinearAnalysis(grid=self.grid,
                                         distributed_slack=self.options.distributed_slack,
                                         correct_values=self.options.correct_values)
        linear_analysis.run()

        # get the contingency branches
        br_idx = linear_analysis.numerical_circuit.branch_data.get_contingency_enabled_indices()

        self.progress_text.emit('Computing branch base loading...')
        Pbus = ts_numeric_circuit.Sbus.real
        rates = ts_numeric_circuit.ContingencyRates.T

        # get flow
        if self.options.use_provided_flows:
            flows = self.options.Pf

            if self.options.Pf is None:
                msg = 'The option to use the provided flows is enabled, but no flows are available'
                self.logger.add_error(msg)
                raise Exception(msg)
        else:
            # compute the base Sf
            flows = linear_analysis.get_flows_time_series(Pbus)  # will be converted to MW internally

        self.progress_text.emit('Computing N-1 loading...')

        for ie, e in enumerate(br_idx):

            self.progress_text.emit('Computing N-1 loading...' + ts_numeric_circuit.branch_names[e])

            # the results are passed by reference and filled inside
            compute_flows_numba(e=e,
                                nt=nt,
                                contingency_branch_idx=br_idx,
                                LODF=linear_analysis.LODF,
                                Flows=flows,
                                rates=rates,
                                overload_count=results.overload_count,
                                max_overload=results.max_overload,
                                worst_flows=results.worst_flows)

            self.progress_signal.emit((ie + 1) / len(br_idx) * 100)

            if self.__cancel__:
                results.relative_frequency = results.overload_count / nt
                results.worst_loading = results.worst_flows / (rates + 1e-9)
                return results

        results.relative_frequency = results.overload_count / nt
        results.worst_loading = results.worst_flows / (rates + 1e-9)
        results.S = Pbus.T

        return results

    def run(self):
        """

        :return:
        """
        start = time.time()
        self.results = self.n_minus_k()

        end = time.time()
        self.elapsed = end - start


if __name__ == '__main__':
    import os
    import pandas as pd
    from GridCal.Engine import FileOpen, SolverType, PowerFlowOptions

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Lynn 5 Bus pv.gridcal'
    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/grid_2_islands.xlsx'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/2869 Pegase.gridcal'
    # fname = os.path.join('..', '..', '..', '..', '..', 'Grids_and_profiles', 'grids', 'IEEE 30 Bus with storage.xlsx')
    # fname = os.path.join('..', '..', '..', '..', '..', 'Grids_and_profiles', 'grids', '2869 Pegase.gridcal')

    main_circuit = FileOpen(fname).open()

    options_ = ContingencyAnalysisOptions()
    simulation = ContingencyAnalysisTimeSeries(grid=main_circuit, options=options_)
    simulation.run()

    print()



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
This file implements a DC-OPF for time series
That means that solves the OPF problem for a complete time series at once
"""

from GridCal.Engine.Core.snapshot_opf_data import SnapshotOpfData
from GridCal.Engine.Simulations.OPF.opf_templates import Opf, MIPSolvers
from GridCal.ThirdParty.pulp import *


def get_objective_function(Pg, Pb, LSlack, FSlack1, FSlack2,
                           cost_g, cost_b, cost_l, cost_br):
    """
    Add the objective function to the problem
    :param Pg: generator LpVars (ng, nt)
    :param Pb: batteries LpVars (nb, nt)
    :param LSlack: Load slack LpVars (nl, nt)
    :param FSlack1: Branch overload slack1 (m, nt)
    :param FSlack2: Branch overload slack2 (m, nt)
    :param cost_g: Cost of the generators (ng, nt)
    :param cost_b: Cost of the batteries (nb, nt)
    :param cost_l: Cost of the loss of load (nl, nt)
    :param cost_br: Cost of the overload (m, nt)
    :return: Nothing, just assign the objective function
    """

    f_obj = lpSum(cost_g * Pg)

    f_obj += lpSum(cost_b * Pb)

    f_obj += lpSum(cost_l * LSlack)

    f_obj += lpSum(cost_br * (FSlack1 + FSlack2))

    return f_obj


def set_fix_generation(problem, Pg, P_fix, enabled_for_dispatch):
    """
    Set the generation fixed at the non dispatchable generators
    :param problem: LP problem instance
    :param Pg: Array of generation variables
    :param P_fix: Array of fixed generation values
    :param enabled_for_dispatch: array of "enables" for dispatching generators
    :return: Nothing
    """

    idx = np.where(enabled_for_dispatch == False)[0]

    lpAddRestrictions2(problem=problem,
                       lhs=Pg[idx],
                       rhs=P_fix[idx],
                       name='fixed_generation',
                       op='=')


def get_power_injections(C_bus_gen, Pg, C_bus_bat, Pb, C_bus_load, PlSlack, QlSlack, Pl, Ql):
    """
    Create the power injections per bus
    :param C_bus_gen: Bus-Generators sparse connectivity matrix (n, ng)
    :param Pg: generator LpVars (ng, nt)
    :param C_bus_bat: Bus-Batteries sparse connectivity matrix (n, nb)
    :param Pb: Batteries LpVars (nb, nt)
    :param C_bus_load: Bus-Load sparse connectivity matrix (n, nl)
    :param PlSlack: Load (real) slack LpVars (nl, nt)
    :param QlSlack: Load (imag) slack LpVars (nl, nt)
    :param Pl: Load values (nl, nt)
    :return: Power injection at the buses (n, nt)
    """

    P = lpDot(C_bus_gen, Pg)

    P += lpDot(C_bus_bat, Pb)

    P -= lpDot(C_bus_load, Pl - PlSlack)

    Q = - lpDot(C_bus_load, Ql - QlSlack)

    return P, Q


def add_ac_nodal_power_balance(numerical_circuit, problem: LpProblem, dvm, dva, P, Q):
    """
    Add the nodal power balance
    :param numerical_circuit: NumericalCircuit instance
    :param problem: LpProblem instance
    :param dva: Voltage angles LpVars (n, nt)
    :param P: Power injection at the buses LpVars (n, nt)
    :return: Nothing, the restrictions are added to the problem
    """

    # do the topological computation
    calculation_inputs = numerical_circuit.split_into_islands()

    nodal_restrictions_P = np.empty(numerical_circuit.nbus, dtype=object)
    nodal_restrictions_Q = np.empty(numerical_circuit.nbus, dtype=object)

    # simulate each island and merge the results
    for i, calc_inpt in enumerate(calculation_inputs):

        if len(calc_inpt.vd) > 0:
            # find the original indices
            bus_original_idx = np.array(calc_inpt.original_bus_idx)

            # re-pack the variables for the island and time interval
            P_island = P[bus_original_idx]  # the sizes already reflect the correct time span
            Q_island = Q[bus_original_idx]  # the sizes already reflect the correct time span
            dva_island = dva[bus_original_idx]  # the sizes already reflect the correct time span
            dvm_island = dvm[bus_original_idx]  # the sizes already reflect the correct time span

            B_island = calc_inpt.Ybus.imag
            G_island = calc_inpt.Ybus.real
            Bs_island = calc_inpt.Yseries.imag
            Gs_island = calc_inpt.Yseries.real

            pqpv = calc_inpt.pqpv
            pq = calc_inpt.pq
            pv = calc_inpt.pv
            vd = calc_inpt.vd
            vdpv = np.r_[vd, pv]
            vdpv.sort()

            # Add nodal real power balance for the non slack nodes
            idx = bus_original_idx[pqpv]
            nodal_restrictions_P[idx] = lpAddRestrictions2(problem=problem,
                                                           lhs=-lpDot(Bs_island[np.ix_(pqpv, pqpv)], dva_island[pqpv])
                                                               + lpDot(G_island[np.ix_(pqpv, pq)], dvm_island[pq]),
                                                           rhs=P_island[pqpv],
                                                           name='Nodal_real_power_balance_pqpv_is' + str(i),
                                                           op='=')

            # Add nodal reactive power balance for the non slack nodes
            idx = bus_original_idx[pq]
            nodal_restrictions_Q[idx] = lpAddRestrictions2(problem=problem,
                                                           lhs=-lpDot(Gs_island[np.ix_(pq, pqpv)], dva_island[pqpv])
                                                               - lpDot(B_island[np.ix_(pq, pq)], dvm_island[pq]),
                                                           rhs=Q_island[pq],
                                                           name='Nodal_imag_power_balance_pqpv_is' + str(i),
                                                           op='=')

            # Add nodal real power balance for the slack nodes
            idx = bus_original_idx[vd]
            nodal_restrictions_P[idx] = lpAddRestrictions2(problem=problem,
                                                           lhs=-lpDot(Bs_island[vd, :], dva_island)
                                                               + lpDot(G_island[vd, :], dvm_island),
                                                           rhs=P_island[vd],
                                                           name='Nodal_real_power_balance_vd_is' + str(i),
                                                           op='=')

            # delta of voltage angles equal to zero for the slack nodes (vd)
            lpAddRestrictions2(problem=problem,
                               lhs=dva_island[vd],
                               rhs=np.zeros_like(dva_island[vd]),
                               name='dVa_vd_zero_is' + str(i),
                               op='=')

            # delta of voltage module equal to zero for the slack and pv nodes (vdpv)
            lpAddRestrictions2(problem=problem,
                               lhs=dvm_island[vdpv],
                               rhs=np.zeros_like(dva_island[vdpv]),
                               name='dVm_vdpv_zero_is' + str(i),
                               op='=')

    return nodal_restrictions_P, nodal_restrictions_Q


def add_branch_loading_restriction(problem: LpProblem,
                                   theta_f, theta_t, Bseries,
                                   Fmax, FSlack1, FSlack2):
    """
    Add the branch loading restrictions
    :param problem: LpProblem instance
    :param theta_f: voltage angles at the "from" side of the branches (m, nt)
    :param theta_t: voltage angles at the "to" side of the branches (m, nt)
    :param Bseries: Array of branch susceptances (m)
    :param Fmax: Array of branch ratings (m, nt)
    :param FSlack1: Array of branch loading slack variables in the from-to sense
    :param FSlack2: Array of branch loading slack variables in the to-from sense
    :return: Nothing
    """

    load_f = Bseries * (theta_f - theta_t)
    load_t = Bseries * (theta_t - theta_f)

    # from-to branch power restriction
    lpAddRestrictions2(problem=problem,
                       lhs=load_f,
                       rhs=np.array(Fmax + FSlack1),  # Fmax + FSlack1
                       name='from_to_branch_rate',
                       op='<=')

    # to-from branch power restriction
    lpAddRestrictions2(problem=problem,
                       lhs=load_t,
                       rhs=np.array(Fmax + FSlack2),  # Fmax + FSlack2
                       name='to_from_branch_rate',
                       op='<=')

    return load_f, load_t


class OpfAc(Opf):

    def __init__(self, numerical_circuit: SnapshotOpfData, solver_type: MIPSolvers = MIPSolvers.CBC):
        """
        DC time series linear optimal power flow
        :param numerical_circuit: NumericalCircuit instance
        """
        Opf.__init__(self, numerical_circuit=numerical_circuit, solver_type=solver_type)

        self.v0 = None
        self.dva = None
        self.dvm = None
        self.Ql = None

    def formulate(self):
        """
        Formulate the AC OPF time series in the non-sequential fashion (all to the solver_type at once)
        :return: PuLP Problem instance
        """
        numerical_circuit = self.numerical_circuit

        # general indices
        n = numerical_circuit.nbus
        m = numerical_circuit.nbr
        ng = numerical_circuit.ngen
        nb = numerical_circuit.nbatt
        nl = numerical_circuit.nload
        Sbase = numerical_circuit.Sbase

        # battery
        Pb_max = numerical_circuit.battery_pmax / Sbase
        Pb_min = numerical_circuit.battery_pmin / Sbase
        cost_b = numerical_circuit.battery_cost

        # generator
        Pg_max = numerical_circuit.generator_pmax / Sbase
        Pg_min = numerical_circuit.generator_pmin / Sbase
        cost_g = numerical_circuit.generator_cost
        P_fix = numerical_circuit.generator_p / Sbase
        enabled_for_dispatch = numerical_circuit.generator_dispatchable

        # load
        Pl = (numerical_circuit.load_active * numerical_circuit.load_s.real) / Sbase
        Ql = (numerical_circuit.load_active * numerical_circuit.load_s.imag) / Sbase
        cost_l = numerical_circuit.load_cost

        # branch
        branch_ratings = numerical_circuit.branch_rates / Sbase
        Bseries = (numerical_circuit.branch_active * (1 / (numerical_circuit.branch_R + 1j * numerical_circuit.branch_X))).imag
        cost_br = numerical_circuit.branch_cost

        # create LP variables
        Pg = lpMakeVars(name='Pg', shape=ng, lower=Pg_min, upper=Pg_max)
        Pb = lpMakeVars(name='Pb', shape=nb, lower=Pb_min, upper=Pb_max)
        load_slack = lpMakeVars(name='LSlack', shape=nl, lower=0, upper=None)
        dva = lpMakeVars(name='dva', shape=n, lower=-3.14, upper=3.14)
        dvm = lpMakeVars(name='dvm', shape=n, lower=0, upper=2)
        theta_f = dva[numerical_circuit.F]
        theta_t = dva[numerical_circuit.T]
        branch_rating_slack1 = lpMakeVars(name='FSlack1', shape=m, lower=0, upper=None)
        branch_rating_slack2 = lpMakeVars(name='FSlack2', shape=m, lower=0, upper=None)

        # declare problem
        problem = LpProblem(name='AC_OPF')

        # add the objective function
        problem += get_objective_function(Pg, Pb, load_slack, branch_rating_slack1, branch_rating_slack2,
                                          cost_g, cost_b, cost_l, cost_br)

        # set the fixed generation values
        set_fix_generation(problem=problem, Pg=Pg, P_fix=P_fix, enabled_for_dispatch=enabled_for_dispatch)

        # compute the power injections per node
        P, Q = get_power_injections(C_bus_gen=numerical_circuit.generator_data.C_bus_gen, Pg=Pg,
                                    C_bus_bat=numerical_circuit.battery_data.C_bus_batt, Pb=Pb,
                                    C_bus_load=numerical_circuit.load_data.C_bus_load,
                                    PlSlack=load_slack, QlSlack=load_slack,
                                    Pl=Pl, Ql=Ql)

        # compute the nodal power balance restrictions
        nodal_restrictions_P, nodal_restrictions_Q = add_ac_nodal_power_balance(numerical_circuit=numerical_circuit,
                                                                                problem=problem,
                                                                                dvm=dvm, dva=dva, P=P, Q=Q)

        # add the branch loading restriction
        load_f, load_t = add_branch_loading_restriction(problem, theta_f, theta_t, Bseries, branch_ratings,
                                                        branch_rating_slack1, branch_rating_slack2)

        # Assign variables to keep
        # transpose them to be in the format of GridCal: time, device
        self.v0 = np.abs(numerical_circuit.Vbus)
        self.dva = dva
        self.dvm = dvm
        self.Pg = Pg
        self.Pb = Pb
        self.Pl = Pl
        self.load_shedding = load_slack
        self.s_from = load_f
        self.s_to = load_t
        self.overloads = branch_rating_slack1 + branch_rating_slack2
        self.rating = branch_ratings
        self.nodal_restrictions = nodal_restrictions_P

        return problem

    def get_voltage(self):
        """
        return the complex voltages (time, device)
        :return: 2D array
        """
        angles = self.extract(self.dva)
        modules = self.v0 + self.extract(self.dvm)
        return modules * np.exp(-1j * angles)


if __name__ == '__main__':
    from GridCal.Engine.basic_structures import BranchImpedanceMode
    from GridCal.Engine.IO.file_handler import FileOpen
    from GridCal.Engine.Core.snapshot_opf_data import compile_snapshot_opf_circuit

    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Lynn 5 Bus pv.gridcal'
    fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/IEEE39_1W.gridcal'
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Lynn 5 Bus pv (2 islands).gridcal'

    main_circuit = FileOpen(fname).open()

    # main_circuit.buses[3].controlled_generators[0].enabled_dispatch = False

    numerical_circuit_ = compile_snapshot_opf_circuit(circuit=main_circuit,
                                                      apply_temperature=False,
                                                      branch_tolerance_mode=BranchImpedanceMode.Specified)

    problem = OpfAc(numerical_circuit=numerical_circuit_)

    print('Solving...')
    status = problem.solve()

    # print("Status:", status)

    v = problem.get_voltage()
    print('Modules\n', np.abs(v))
    print('Angles\n', np.angle(v))

    l = problem.get_loading()
    print('Branch loading\n', l)

    g = problem.get_generator_power()
    print('Gen power\n', g)

    pr = problem.get_shadow_prices()
    print('Nodal prices \n', pr)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'gui.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from .icons_rc import *

class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        if not AboutDialog.objectName():
            AboutDialog.setObjectName(u"AboutDialog")
        AboutDialog.resize(462, 367)
        icon = QIcon()
        icon.addFile(u":/Icons/icons/GridCal_icon.svg", QSize(), QIcon.Normal, QIcon.Off)
        AboutDialog.setWindowIcon(icon)
        self.verticalLayout_2 = QVBoxLayout(AboutDialog)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.tabWidget = QTabWidget(AboutDialog)
        self.tabWidget.setObjectName(u"tabWidget")
        self.tab = QWidget()
        self.tab.setObjectName(u"tab")
        self.gridLayout = QGridLayout(self.tab)
        self.gridLayout.setObjectName(u"gridLayout")
        self.mainLabel = QLabel(self.tab)
        self.mainLabel.setObjectName(u"mainLabel")
        self.mainLabel.setLayoutDirection(Qt.LeftToRight)
        self.mainLabel.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.mainLabel.setWordWrap(True)
        self.mainLabel.setOpenExternalLinks(True)
        self.mainLabel.setTextInteractionFlags(Qt.TextBrowserInteraction)

        self.gridLayout.addWidget(self.mainLabel, 0, 1, 2, 2)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 1, 0, 1, 1)

        self.label = QLabel(self.tab)
        self.label.setObjectName(u"label")
        self.label.setMinimumSize(QSize(48, 48))
        self.label.setMaximumSize(QSize(48, 48))
        self.label.setPixmap(QPixmap(u":/Icons/icons/GridCal_icon.svg"))
        self.label.setScaledContents(True)

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.versionLabel = QLabel(self.tab)
        self.versionLabel.setObjectName(u"versionLabel")
        self.versionLabel.setOpenExternalLinks(True)
        self.versionLabel.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByMouse)

        self.gridLayout.addWidget(self.versionLabel, 3, 1, 1, 1)

        self.copyrightLabel = QLabel(self.tab)
        self.copyrightLabel.setObjectName(u"copyrightLabel")
        self.copyrightLabel.setOpenExternalLinks(True)
        self.copyrightLabel.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByMouse)

        self.gridLayout.addWidget(self.copyrightLabel, 4, 1, 1, 1)

        self.tabWidget.addTab(self.tab, "")
        self.tab_2 = QWidget()
        self.tab_2.setObjectName(u"tab_2")
        self.verticalLayout = QVBoxLayout(self.tab_2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.contributorsLabel = QLabel(self.tab_2)
        self.contributorsLabel.setObjectName(u"contributorsLabel")
        self.contributorsLabel.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)

        self.verticalLayout.addWidget(self.contributorsLabel)

        self.tabWidget.addTab(self.tab_2, "")
        self.tab_3 = QWidget()
        self.tab_3.setObjectName(u"tab_3")
        self.gridLayout_2 = QGridLayout(self.tab_3)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.updateLabel = QLabel(self.tab_3)
        self.updateLabel.setObjectName(u"updateLabel")
        self.updateLabel.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)
        self.updateLabel.setWordWrap(True)
        self.updateLabel.setTextInteractionFlags(Qt.LinksAccessibleByMouse|Qt.TextSelectableByMouse)

        self.gridLayout_2.addWidget(self.updateLabel, 0, 1, 1, 1)

        self.updateButton = QPushButton(self.tab_3)
        self.updateButton.setObjectName(u"updateButton")
        self.updateButton.setMaximumSize(QSize(80, 16777215))

        self.gridLayout_2.addWidget(self.updateButton, 0, 0, 1, 1)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout_2.addItem(self.verticalSpacer_2, 1, 0, 1, 1)

        self.tabWidget.addTab(self.tab_3, "")
        self.tab_4 = QWidget()
        self.tab_4.setObjectName(u"tab_4")
        self.verticalLayout_3 = QVBoxLayout(self.tab_4)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.licenseTextEdit = QTextEdit(self.tab_4)
        self.licenseTextEdit.setObjectName(u"licenseTextEdit")

        self.verticalLayout_3.addWidget(self.licenseTextEdit)

        self.tabWidget.addTab(self.tab_4, "")

        self.verticalLayout_2.addWidget(self.tabWidget)


        self.retranslateUi(AboutDialog)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(AboutDialog)
    # setupUi

    def retranslateUi(self, AboutDialog):
        AboutDialog.setWindowTitle(QCoreApplication.translate("AboutDialog", u"About GridCal", None))
        self.mainLabel.setText(QCoreApplication.translate("AboutDialog", u"<html><head/><body><p align=\"justify\"><span style=\" font-weight:600;\">GridCal</span> has been carefully crafted since 2015 to serve as a platform for research and consultancy. </p><p align=\"justify\">Visit <a href=\"https://gridcal.org\"><span style=\" text-decoration: underline; color:#0000ff;\">https://gridcal.org</span></a> for more details.</p><p align=\"justify\">This program comes with absolutelly no warranty. This is free software, and you are welcome to redistribute it under the conditions set by the license. GridCal is licensed under the lesser GNU general public license version 3. See the license file for more details.</p><p align=\"justify\">The source of GridCal can be found at: <a href=\"https://github.com/SanPen/GridCal\"><span style=\" text-decoration: underline; color:#0000ff;\">https://github.com/SanPen/GridCal</span></a></p></body></html>", None))
        self.label.setText("")
        self.versionLabel.setText(QCoreApplication.translate("AboutDialog", u"version", None))
        self.copyrightLabel.setText(QCoreApplication.translate("AboutDialog", u"Copyright", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), QCoreApplication.translate("AboutDialog", u"About", None))
        self.contributorsLabel.setText(QCoreApplication.translate("AboutDialog", u"TextLabel", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), QCoreApplication.translate("AboutDialog", u"Contributors", None))
        self.updateLabel.setText(QCoreApplication.translate("AboutDialog", u"TextLabel", None))
        self.updateButton.setText(QCoreApplication.translate("AboutDialog", u"Update", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_3), QCoreApplication.translate("AboutDialog", u"Update", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_4), QCoreApplication.translate("AboutDialog", u"License", None))
    # retranslateUi




import numpy as np


def short_circuit_3p(bus_idx, Zbus, Vbus, Zf, baseMVA):
    """
    Executes a 3-phase balanced short circuit study
    Args:
        bus_idx: Index of the bus at which the short circuit is being studied
        Zbus: Inverse of the admittance matrix
        Vbus: Voltages of the buses in the steady state
        Zf: Fault impedance array

    Returns: Voltages after the short circuit (p.u.), Short circuit power in MVA

    """
    n = len(Vbus)
    Z = Zbus[bus_idx, :][:, bus_idx]  # Z_B in documentation

    # Voltage Source Contribution
    I_kI = np.zeros(n, dtype=complex)
    Z.flat[::len(bus_idx) + 1] += Zf[bus_idx]  # add Zf to diagonals of Z_B
    I_kI[bus_idx] = -1 * np.linalg.solve(Z, Vbus[bus_idx])

    # Current source contribution
    # I_kII = -1 * Zbus.dot(I_kC / Z[elm_idx])

    # Total current contribution
    # I_k = I_kI + I_kII
    I_k = I_kI

    # voltage increment due to these currents
    incV = Zbus.dot(I_k)

    V = Vbus + incV

    # Short circuit power in MVA
    SCC = -I_k * Vbus * baseMVA

    return V, SCC


from GridCal.Engine.Core.topology import find_islands, get_adjacency_matrix, get_elements_of_the_island
from GridCal.Engine.Core.snapshot_pf_data import SnapshotData, compile_snapshot_circuit
from GridCal.Engine.Core.snapshot_opf_data import SnapshotOpfData, compile_snapshot_opf_circuit
from GridCal.Engine.Core.time_series_pf_data import TimeCircuit, compile_time_circuit
from GridCal.Engine.Core.time_series_opf_data import OpfTimeCircuit, compile_opf_time_circuit
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Core.admittance_matrices import Admittance, compute_linear_admittances, compute_admittances, compute_connectivity




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import sys
from typing import List, Dict, Tuple
from uuid import getnode as get_mac, uuid4
from datetime import timedelta, datetime
import networkx as nx
from scipy.sparse import csc_matrix, lil_matrix
# from GridCal.Gui.GeneralDialogues import *
from GridCal.Engine.Devices import *
from GridCal.Engine.Devices.editable_device import DeviceType


def get_system_user():
    """
    Get the system mac + user name
    :return: string with the system mac address and the current user
    """

    # get the proper function to find the user depending on the platform
    if 'USERNAME' in os.environ:
        user = os.environ["USERNAME"]
    elif 'USER' in os.environ:
        user = os.environ["USER"]
    else:
        user = ''

    try:
        mac = get_mac()
    except:
        mac = ''

    return str(mac) + ':' + user


class MultiCircuit:
    """
    The concept of circuit should be easy enough to understand. It represents a set of
    nodes (:ref:`buses<Bus>`) and :ref:`branches<Branch>` (lines, transformers or other
    impedances).

    The :ref:`MultiCircuit<multicircuit>` class is the main object in **GridCal**. It
    represents a circuit that may contain islands. It is important to understand that a
    circuit split in two or more islands cannot be simulated as is, because the
    admittance matrix would be singular. The solution to this is to split the circuit
    in island-circuits. Therefore :ref:`MultiCircuit<multicircuit>` identifies the
    islands and creates individual **Circuit** objects for each of them.

    **GridCal** uses an object oriented approach for the data management. This allows
    to group the data in a smart way. In **GridCal** there are only two types of object
    directly declared in a **Circuit** or :ref:`MultiCircuit<multicircuit>` object.
    These are the :ref:`Bus<bus>` and the :ref:`Branch<branch>`. The branches connect
    the buses and the buses contain all the other possible devices like loads,
    generators, batteries, etc. This simplifies enormously the management of element
    when adding, associating and deleting.

    .. code:: ipython3

        from GridCal.Engine.Core.multi_circuit import MultiCircuit
        grid = MultiCircuit(name="My grid")

    """

    def __init__(self, name='', Sbase=100, fbase=50.0, idtag=None):
        """
        class constructor
        :param name: name of the circuit
        :param Sbase: base power in MVA
        :param fbase: base frequency in Hz
        :param idtag: unique identifier
        """

        self.name = name

        if idtag is None:
            self.idtag = uuid4().hex
        else:
            self.idtag = idtag

        self.comments = ''

        # this is a number that serves
        self.model_version = 2

        # user mane
        self.user_name = get_system_user()

        # Base power (MVA)
        self.Sbase = Sbase

        # Base frequency in Hz
        self.fBase = fbase

        # Should be able to accept Branches, Lines and Transformers alike
        # self.branches = list()

        self.lines: List[Line] = list()

        self.dc_lines: List[DcLine] = list()

        self.transformers2w: List[Transformer2W] = list()

        self.hvdc_lines: List[HvdcLine] = list()

        self.vsc_devices: List[VSC] = list()

        self.upfc_devices: List[UPFC] = list()

        self.switch_devices: List[Switch] = list()

        self.transformers3w: List[Transformer3W] = list()

        # array of branch indices in the master circuit
        self.branch_original_idx = list()

        # Should accept buses
        self.buses: List[Bus] = list()

        # array of bus indices in the master circuit
        self.bus_original_idx = list()

        # Dictionary relating the bus object to its index. Updated upon compilation
        self.buses_dict: Dict[Bus, int] = dict()

        # List of overhead line objects
        self.overhead_line_types: List[Tower] = list()

        # list of wire types
        self.wire_types: List[Wire] = list()

        # underground cable lines
        self.underground_cable_types: List[UndergroundLineType] = list()

        # sequence modelled lines
        self.sequence_line_types: List[SequenceLineType] = list()

        # List of transformer types
        self.transformer_types: List[TransformerType] = list()

        # list of substations
        self.default_substation = Substation('Default substation')
        self.substations: List[Substation] = [self.default_substation]

        # list of areas
        self.default_area = Area('Default area')
        self.areas: List[Area] = [self.default_area]

        # list of zones
        self.default_zone = Zone('Default zone')
        self.zones: List[Zone] = [self.default_zone]

        # list of countries
        self.default_country = Country('Default country')
        self.countries: List[Country] = [self.default_country]

        # logger of events
        self.logger = Logger()

        # Object with the necessary inputs for a power flow study
        self.numerical_circuit = None

        # Bus-Branch graph
        self.graph = None

        # dictionary of bus objects -> bus indices
        self.bus_dictionary = dict()

        # dictionary of branch objects -> branch indices
        self.branch_dictionary = dict()

        # names of the buses
        self.bus_names = None

        # names of the branches
        self.branch_names = None

        # master time profile
        self.time_profile = None

        # objects with profiles
        self.objects_with_profiles = [Bus(),
                                      Load(),
                                      StaticGenerator(),
                                      ExternalGrid(),
                                      Generator(),
                                      Battery(),
                                      Shunt(),
                                      Line(None, None),
                                      DcLine(None, None),
                                      Transformer2W(None, None),
                                      Transformer3W(),
                                      HvdcLine(None, None),
                                      VSC(None, None),
                                      UPFC(None, None),
                                      Substation(),
                                      Zone(),
                                      Area(),
                                      Country()]

        # dictionary of profile magnitudes per object
        self.profile_magnitudes = dict()

        self.device_type_name_dict = dict()

        '''
        self.type_name = 'Shunt'

        self.properties_with_profile = ['Y']
        '''
        for dev in self.objects_with_profiles:
            if dev.properties_with_profile is not None:
                profile_attr = list(dev.properties_with_profile.keys())
                profile_types = [dev.editable_headers[attr].tpe for attr in profile_attr]
                self.profile_magnitudes[dev.device_type.value] = (profile_attr, profile_types)
                self.device_type_name_dict[dev.device_type.value] = dev.device_type

    def __str__(self):
        return str(self.name)

    @property
    def has_time_series(self):
        return self.time_profile is not None

    def get_zones(self):
        return self.zones

    def get_zone_number(self):
        """
        Get number of areas
        :return: number of areas
        """
        return len(self.zones)

    def get_areas(self):
        return self.areas

    def get_area_number(self):
        """
        Get number of areas
        :return: number of areas
        """
        return len(self.areas)

    def get_substations(self):
        return self.substations

    def get_substation_number(self):
        """
        Get number of areas
        :return: number of areas
        """
        return len(self.substations)

    def get_bus_number(self):
        """
        Return the number of buses
        :return: number
        """
        return len(self.buses)

    def get_branch_lists_wo_hvdc(self):
        """
        GEt list of the branch lists
        :return:
        """
        return [self.lines, self.transformers2w,  self.vsc_devices, self.dc_lines, self.upfc_devices]

    def get_branch_lists(self):
        """
        GEt list of the branch lists
        :return:
        """
        return self.get_branch_lists_wo_hvdc() + [self.hvdc_lines]

    def get_branch_number(self):
        """
        return the number of branches (of all types)
        :return: number
        """
        m = 0
        for branch_list in self.get_branch_lists():
            m += len(branch_list)
        return m

    def get_branch_number_wo_hvdc(self):
        """
        return the number of branches (of all types)
        :return: number
        """
        m = 0
        for branch_list in self.get_branch_lists_wo_hvdc():
            m += len(branch_list)
        return m

    def get_time_number(self):
        """
        Return the number of buses
        :return: number
        """
        if self.time_profile is not None:
            return len(self.time_profile)
        else:
            return 0

    def get_dimensions(self):
        """
        Get the three dimensions of the circuit: number of buses, number of branches, number of time steps
        :return: (nbus, nbranch, ntime)
        """
        return self.get_bus_number(), self.get_branch_number(), self.get_time_number()

    def clear(self):
        """
        Clear the multi-circuit (remove the bus and branch objects)
        """
        # Should be able to accept Branches, Lines and Transformers alike
        self.lines = list()
        self.dc_lines = list()
        self.transformers2w = list()
        self.hvdc_lines = list()
        self.vsc_devices = list()
        self.upfc_devices = list()

        # array of branch indices in the master circuit
        self.branch_original_idx = list()

        # Should accept buses
        self.buses = list()

        # array of bus indices in the master circuit
        self.bus_original_idx = list()

        # Dictionary relating the bus object to its index. Updated upon compilation
        self.buses_dict = dict()

        # List of overhead line objects
        self.overhead_line_types = list()

        # list of wire types
        self.wire_types = list()

        # underground cable lines
        self.underground_cable_types = list()

        # sequence modelled lines
        self.sequence_line_types = list()

        # List of transformer types
        self.transformer_types = list()

        # Object with the necessary inputs for a power flow study
        self.numerical_circuit = None

        # Bus-Branch graph
        self.graph = None

        self.bus_dictionary = dict()

        self.branch_dictionary = dict()

        self.bus_names = None

        self.branch_names = None

        self.time_profile = None

    def get_buses(self):
        return self.buses

    def get_bus_names(self):
        return [e.name for e in self.buses]

    def get_branches_wo_hvdc(self):
        """
        Return all the branch objects
        :return: lines + transformers 2w + hvdc
        """
        return self.lines + self.dc_lines + self.transformers2w + self.vsc_devices + self.upfc_devices + self.switch_devices

    def get_branches_wo_hvdc_names(self):
        return [e.name for e in self.get_branches_wo_hvdc()]

    def get_branches(self):
        """
        Return all the branch objects
        :return: lines + transformers 2w + hvdc
        """
        return self.get_branches_wo_hvdc() + self.hvdc_lines

    def get_lines(self) -> List[Line]:
        return self.lines

    def get_transformers2w(self) -> List[Transformer2W]:
        return self.transformers2w

    def get_transformers2w_number(self) -> int:
        return len(self.transformers2w)

    def get_vsc(self) -> List[VSC]:
        return self.vsc_devices

    def get_dc_lines(self) -> List[DcLine]:
        return self.dc_lines

    def get_upfc(self) -> List[UPFC]:
        return self.upfc_devices

    def get_switches(self) -> List[Switch]:
        return self.switch_devices

    def get_hvdc(self) -> List[HvdcLine]:
        return self.hvdc_lines

    def get_hvdc_number(self) -> int:
        return len(self.hvdc_lines)

    def get_loads(self) -> List[Load]:
        """
        Returns a list of :ref:`Load<load>` objects in the grid.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.loads:
                elm.bus = bus
            lst = lst + bus.loads
        return lst

    def get_load_names(self):
        """
        Returns a list of :ref:`Load<load>` names.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.loads:
                lst.append(elm.name)
        return np.array(lst)

    def get_external_grids(self) -> List[ExternalGrid]:
        """
        Returns a list of :ref:`ExternalGrid<external_grid>` objects in the grid.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.external_grids:
                elm.bus = bus
            lst = lst + bus.external_grids
        return lst

    def get_external_grid_names(self):
        """
        Returns a list of :ref:`ExternalGrid<external_grid>` names.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.external_grids:
                lst.append(elm.name)
        return np.array(lst)

    def get_static_generators(self) -> List[StaticGenerator]:
        """
        Returns a list of :ref:`StaticGenerator<static_generator>` objects in the grid.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.static_generators:
                elm.bus = bus
            lst = lst + bus.static_generators
        return lst

    def get_static_generators_names(self):
        """
        Returns a list of :ref:`StaticGenerator<static_generator>` names.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.static_generators:
                lst.append(elm.name)
        return np.array(lst)

    def get_shunts(self) -> List[Shunt]:
        """
        Returns a list of :ref:`Shunt<shunt>` objects in the grid.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.shunts:
                elm.bus = bus
            lst = lst + bus.shunts
        return lst

    def get_shunt_names(self):
        """
        Returns a list of :ref:`Shunt<shunt>` names.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.shunts:
                lst.append(elm.name)
        return np.array(lst)

    def get_generators(self) -> List[Generator]:
        """
        Returns a list of :ref:`Generator<generator>` objects in the grid.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.controlled_generators:
                elm.bus = bus
            lst = lst + bus.controlled_generators
        return lst

    def get_controlled_generator_names(self):
        """
        Returns a list of :ref:`Generator<generator>` names.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.controlled_generators:
                lst.append(elm.name)
        return np.array(lst)

    def get_batteries(self) -> List[Battery]:
        """
        Returns a list of :ref:`Battery<battery>` objects in the grid.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.batteries:
                elm.bus = bus
            lst = lst + bus.batteries
        return lst

    def get_battery_names(self):
        """
        Returns a list of :ref:`Battery<battery>` names.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.batteries:
                lst.append(elm.name)
        return np.array(lst)

    def get_battery_capacities(self):
        """
        Returns a list of :ref:`Battery<battery>` capacities.
        """
        lst = list()
        for bus in self.buses:
            for elm in bus.batteries:
                lst.append(elm.Enom)
        return np.array(lst)

    def get_elements_by_type(self, element_type: DeviceType):
        """
        Get set of elements and their parent nodes
        :param element_type: DeviceTYpe instance
        :return: List of elements, it raises an exception if the elements are unknown
        """

        if element_type == DeviceType.LoadDevice:
            return self.get_loads()

        elif element_type == DeviceType.StaticGeneratorDevice:
            return self.get_static_generators()

        elif element_type == DeviceType.GeneratorDevice:
            return self.get_generators()

        elif element_type == DeviceType.BatteryDevice:
            return self.get_batteries()

        elif element_type == DeviceType.ShuntDevice:
            return self.get_shunts()

        elif element_type == DeviceType.ExternalGridDevice:
            return self.get_external_grids()

        elif element_type == DeviceType.LineDevice:
            return self.lines

        elif element_type == DeviceType.Transformer2WDevice:
            return self.transformers2w

        elif element_type == DeviceType.HVDCLineDevice:
            return self.hvdc_lines

        elif element_type == DeviceType.UpfcDevice:
            return self.upfc_devices

        elif element_type == DeviceType.VscDevice:
            return self.vsc_devices

        elif element_type == DeviceType.BusDevice:
            return self.buses

        elif element_type == DeviceType.TowerDevice:
            return self.overhead_line_types

        elif element_type == DeviceType.TransformerTypeDevice:
            return self.transformer_types

        elif element_type == DeviceType.UnderGroundLineDevice:
            return self.underground_cable_types

        elif element_type == DeviceType.SequenceLineDevice:
            return self.sequence_line_types

        elif element_type == DeviceType.WireDevice:
            return self.wire_types

        elif element_type == DeviceType.DCLineDevice:
            return self.dc_lines

        elif element_type == DeviceType.SubstationDevice:
            return self.substations

        elif element_type == DeviceType.AreaDevice:
            return self.areas

        elif element_type == DeviceType.ZoneDevice:
            return self.zones

        elif element_type == DeviceType.CountryDevice:
            return self.countries

        else:
            raise Exception('Element type not understood ' + str(element_type))

    def get_node_elements_by_type2(self, element_type: DeviceType):
        """
        Get set of elements and their parent nodes
        :param element_type: DeviceTYpe instance
        :return: List of elements, it raises an exception if the elements are unknown
        """

        if element_type == DeviceType.LoadDevice:
            return self.get_loads()

        elif element_type == DeviceType.StaticGeneratorDevice:
            return self.get_static_generators()

        elif element_type == DeviceType.GeneratorDevice:
            return self.get_generators()

        elif element_type == DeviceType.BatteryDevice:
            return self.get_batteries()

        elif element_type == DeviceType.ShuntDevice:
            return self.get_shunts()

        elif element_type == DeviceType.ExternalGridDevice:
            return self.get_external_grids()

        elif element_type == DeviceType.SubstationDevice:
            return [x.substation for x in self.buses]

        elif element_type == DeviceType.AreaDevice:
            return [x.area for x in self.buses]

        elif element_type == DeviceType.ZoneDevice:
            return [x.zone for x in self.buses]

        elif element_type == DeviceType.CountryDevice:
            return [x.country for x in self.buses]

        else:
            raise Exception('Element type not understood ' + str(element_type))

    def apply_lp_profiles(self, results: "OptimalPowerFlowTimeSeriesResults"):
        """
        Apply the LP results as device profiles.
        """
        generators = self.get_generators()
        for i, gen in enumerate(generators):
            gen.P_prof = results.generator_power[:, i]

        loads = self.get_loads()
        for i, gen in enumerate(loads):
            gen.P_prof -= results.load_shedding[:, i]

        # TODO: implement more devices

    def copy(self):
        """
        Returns a deep (true) copy of this circuit.
        """

        cpy = MultiCircuit()

        cpy.name = self.name

        bus_dict = dict()
        for bus in self.buses:
            bus_cpy = bus.copy()
            bus_dict[bus] = bus_cpy
            cpy.buses.append(bus_cpy)

        for branch in self.lines:
            cpy.lines.append(branch.copy(bus_dict))

        for branch in self.transformers2w:
            cpy.transformers2w.append(branch.copy(bus_dict))

        for branch in self.hvdc_lines:
            cpy.hvdc_lines.append(branch.copy(bus_dict))

        for branch in self.vsc_devices:
            cpy.vsc_devices.append(branch.copy(bus_dict))

        cpy.Sbase = self.Sbase

        cpy.branch_original_idx = self.branch_original_idx.copy()

        cpy.bus_original_idx = self.bus_original_idx.copy()

        cpy.time_profile = self.time_profile.copy()

        # cpy.numerical_circuit = self.numerical_circuit.copy()

        return cpy

    def get_catalogue_dict(self, branches_only=False):
        """
        Returns a dictionary with the catalogue types and the associated list of objects.

        Arguments:

            **branches_only** (bool, False): Only branch types
        """
        # 'Wires', 'Overhead lines', 'Underground lines', 'Sequence lines', 'Transformers'

        if branches_only:

            catalogue_dict = {'Overhead lines': self.overhead_line_types,
                              'Transformers': self.transformer_types,
                              'Underground lines': self.underground_cable_types,
                              'Sequence lines': self.sequence_line_types}
        else:
            catalogue_dict = {'Wires': self.wire_types,
                              'Overhead lines': self.overhead_line_types,
                              'Underground lines': self.underground_cable_types,
                              'Sequence lines': self.sequence_line_types,
                              'Transformers': self.transformer_types}

        return catalogue_dict

    def get_catalogue_dict_by_name(self, type_class: str = None):
        """
        Get the catalogue elements by name
        :param type_class:
        :return:
        """
        d = dict()

        # ['Wires', 'Overhead lines', 'Underground lines', 'Sequence lines', 'Transformers']

        if type_class is None:
            tpes = [self.overhead_line_types,
                    self.underground_cable_types,
                    self.wire_types,
                    self.transformer_types,
                    self.sequence_line_types]
            name_prop = ''

        elif type_class == 'Wires':
            tpes = self.wire_types
            name_prop = 'name'

        elif type_class == 'Overhead lines':
            tpes = self.overhead_line_types
            name_prop = 'name'

        elif type_class == 'Underground lines':
            tpes = self.underground_cable_types
            name_prop = 'name'

        elif type_class == 'Sequence lines':
            tpes = self.sequence_line_types
            name_prop = 'name'

        elif type_class == 'Transformers':
            tpes = self.transformer_types
            name_prop = 'name'

        else:
            tpes = list()
            name_prop = 'name'

        # make dictionary
        for tpe in tpes:
            d[getattr(tpe, name_prop)] = tpe

        return d

    def get_properties_dict(self):
        """
        Returns a JSON dictionary of the :ref:`MultiCircuit<multicircuit>` instance
        with the following values: id, type, phases, name, Sbase, comments.

        Arguments:

            **id**: Arbitrary identifier
        """
        d = {'id': self.idtag,
             'phases': 'ps',
             'name': self.name,
             'sbase': self.Sbase,
             'fbase': self.fBase,
             'model_version': self.model_version,
             'user_name': self.user_name,
             'comments': self.comments,
             }

        return d

    def get_units_dict(self):
        """
        """
        return {'time': 'Milliseconds since 1/1/1970 (Unix time in ms)'}

    def get_profiles_dict(self):
        """
        """
        if self.time_profile is not None:
            # recommended way to get the unix datetimes
            arr = (self.time_profile - pd.Timestamp("1970-01-01")) // pd.Timedelta("1s")
            # t = (self.time_profile.array.astype(int) * 1e-9).tolist()  # UNIX time in seconds
            t = arr.tolist()
        else:
            t = list()
        return {'time': t}

    def assign_circuit(self, circ: "MultiCircuit"):
        """
        Assign a circuit object to this object.
        :param circ: Another Circuit instance
        :return:
        """
        self.buses = circ.buses

        self.lines = circ.lines
        self.transformers2w = circ.transformers2w
        self.hvdc_lines = circ.hvdc_lines
        self.vsc_devices = circ.vsc_devices

        self.name = circ.name
        self.Sbase = circ.Sbase
        self.fBase = circ.fBase

        self.sequence_line_types = list(set(self.sequence_line_types + circ.sequence_line_types))
        self.wire_types = list(set(self.wire_types + circ.wire_types))
        self.overhead_line_types = list(set(self.overhead_line_types + circ.overhead_line_types))
        self.underground_cable_types = list(set(self.underground_cable_types + circ.underground_cable_types))
        self.sequence_line_types = list(set(self.sequence_line_types + circ.sequence_line_types))
        self.transformer_types = list(set(self.transformer_types + circ.transformer_types))

    def build_graph(self):
        """
        Returns a networkx DiGraph object of the grid.
        """
        self.graph = nx.DiGraph()

        self.bus_dictionary = {bus.idtag: i for i, bus in enumerate(self.buses)}

        for branch_list in self.get_branch_lists():
            for branch in branch_list:
                f = self.bus_dictionary[branch.bus_from.idtag]
                t = self.bus_dictionary[branch.bus_to.idtag]
                self.graph.add_edge(f, t)

        return self.graph

    def build_graph_real_power_flow(self, current_flow):
        """
        Returns a networkx DiGraph object of the grid.

        Arguments:

            **current_flow** (list): power_flow.results.If object
        """
        self.graph_real_power_flow = nx.DiGraph()

        current_flow_direction = np.real(current_flow) > 0

        for branch_list in self.get_branch_lists():
            for direction, branch in zip(current_flow_direction, branch_list):
                f = self.bus_dictionary[branch.bus_from.idtag]
                t = self.bus_dictionary[branch.bus_to.idtag]
                if direction:
                    self.graph_real_power_flow.add_edge(f, t)
                else:
                    self.graph_real_power_flow.add_edge(t, f)
        return self.graph_real_power_flow

    def create_profiles(self, steps, step_length, step_unit, time_base: datetime = datetime.now()):
        """
        Set the default profiles in all the objects enabled to have profiles.

        Arguments:

            **steps** (int): Number of time steps

            **step_length** (int): Time length (1, 2, 15, ...)

            **step_unit** (str): Unit of the time step ("h", "m" or "s")

            **time_base** (datetime, datetime.now()): Date to start from
        """

        index = np.empty(steps, dtype=object)
        for i in range(steps):
            if step_unit == 'h':
                index[i] = time_base + timedelta(hours=i * step_length)
            elif step_unit == 'm':
                index[i] = time_base + timedelta(minutes=i * step_length)
            elif step_unit == 's':
                index[i] = time_base + timedelta(seconds=i * step_length)

        index = pd.DatetimeIndex(index)

        self.format_profiles(index)

    def format_profiles(self, index):
        """
        Format the pandas profiles in place using a time index.

        Arguments:

            **index**: Time profile
        """

        self.time_profile = pd.to_datetime(index, dayfirst=True)

        for elm in self.buses:
            elm.create_profiles(index)

        for branch_list in self.get_branch_lists():
            for elm in branch_list:
                elm.create_profiles(index)

    def set_time_profile(self, unix_data):

        # try parse
        try:
            self.time_profile = pd.to_datetime(np.array(unix_data), unit='s', origin='unix')
        except Exception as e:
            # it may come in nanoseconds instead of seconds...
            self.time_profile = pd.to_datetime(np.array(unix_data) / 1e9, unit='s', origin='unix')

        for elm in self.buses:
            elm.create_profiles(self.time_profile)

        for branch_list in self.get_branch_lists():
            for elm in branch_list:
                elm.create_profiles(self.time_profile)

    def ensure_profiles_exist(self):
        """
        Format the pandas profiles in place using a time index.
        """
        if self.time_profile is None:
            raise Exception('Cannot ensure profiles existence without a time index. Try format_profiles instead')

        for elm in self.buses:
            elm.ensure_profiles_exist(self.time_profile)

        for branch_list in self.get_branch_lists():
            for elm in branch_list:
                elm.ensure_profiles_exist(self.time_profile)

    def get_bus_dict(self):
        """
        Return dictionary of buses
        :return: dictionary of buses {name:object}
        """
        return {b.name: b for b in self.buses}

    def get_bus_index_dict(self):
        """
        Return dictionary of buses
        :return: dictionary of buses {name:object}
        """
        return {b: i for i, b in enumerate(self.buses)}

    def add_bus(self, obj: Bus):
        """
        Add a :ref:`Bus<bus>` object to the grid.

        Arguments:

            **obj** (:ref:`Bus<bus>`): :ref:`Bus<bus>` object
        """
        if self.time_profile is not None:
            obj.create_profiles(self.time_profile)

        if obj.substation is None:
            obj.substation = self.default_substation

        if obj.zone is None:
            obj.zone = self.default_zone

        if obj.area is None:
            obj.area = self.default_area

        if obj.country is None:
            obj.country = self.default_country

        self.buses.append(obj)

    def delete_bus(self, obj: Bus, ask=True):
        """
        Delete a :ref:`Bus<bus>` object from the grid.

        Arguments:

            **obj** (:ref:`Bus<bus>`): :ref:`Bus<bus>` object
        """

        # remove associated branches in reverse order
        for branch_list in self.get_branch_lists():
            for i in range(len(branch_list) - 1, -1, -1):
                if branch_list[i].bus_from == obj or branch_list[i].bus_to == obj:
                    if branch_list[i].graphic_obj is not None:
                        branch_list[i].graphic_obj.remove(ask=ask)
                    else:
                        self.delete_branch(branch_list[i])

        # remove the bus itself
        if obj in self.buses:
            print('Deleted', obj.name)
            self.buses.remove(obj)

    def add_line(self, obj: Line):
        """
        Add a line object
        :param obj: Line instance
        """

        if self.time_profile is not None:
            obj.create_profiles(self.time_profile)
        self.lines.append(obj)

    def add_dc_line(self, obj: DcLine):
        """
        Add a line object
        :param obj: Line instance
        """

        if self.time_profile is not None:
            obj.create_profiles(self.time_profile)
        self.dc_lines.append(obj)

    def add_transformer2w(self, obj: Transformer2W):
        """
        Add a transformer object
        :param obj: Transformer2W instance
        """

        if self.time_profile is not None:
            obj.create_profiles(self.time_profile)
        self.transformers2w.append(obj)

    def add_transformer3w(self, obj: Transformer3W):
        """
        Add a transformer object
        :param obj: Transformer3W instance
        """

        if self.time_profile is not None:
            obj.create_profiles(self.time_profile)
        self.transformers3w.append(obj)

    def add_hvdc(self, obj: HvdcLine):
        """
        Add a hvdc line object
        :param obj: HvdcLine instance
        """

        if self.time_profile is not None:
            obj.create_profiles(self.time_profile)
        self.hvdc_lines.append(obj)

    def add_vsc(self, obj: VSC):
        """
        Add a hvdc line object
        :param obj: HvdcLine instance
        """

        if self.time_profile is not None:
            obj.create_profiles(self.time_profile)
        self.vsc_devices.append(obj)

    def add_upfc(self, obj: UPFC):
        """
        Add a UPFC object
        :param obj: UPFC instance
        """

        if self.time_profile is not None:
            obj.create_profiles(self.time_profile)
        self.upfc_devices.append(obj)

    def add_switch(self, obj: Switch):
        """
        Add a Switch object
        :param obj: Switch instance
        """

        if self.time_profile is not None:
            obj.create_profiles(self.time_profile)
        self.switch_devices.append(obj)

    def add_branch(self, obj):
        """
        Add a :ref:`Branch<branch>` object to the grid.

        Arguments:

            **obj** (:ref:`Branch<branch>`): :ref:`Branch<branch>` object
        """

        if obj.device_type == DeviceType.LineDevice:
            self.add_line(obj)

        elif obj.device_type == DeviceType.DCLineDevice:
            self.add_dc_line(obj)

        elif obj.device_type == DeviceType.Transformer2WDevice:
            self.add_transformer2w(obj)

        elif obj.device_type == DeviceType.HVDCLineDevice:
            self.add_hvdc(obj)

        elif obj.device_type == DeviceType.VscDevice:
            self.add_vsc(obj)

        elif obj.device_type == DeviceType.UpfcDevice:
            self.add_upfc(obj)

        elif obj.device_type == DeviceType.SwitchDevice:
            self.add_switch(obj)

        elif obj.device_type == DeviceType.BranchDevice:
            # we need to convert it :D
            if obj.branch_type == BranchType.Line or obj.branch_type == BranchType.Transformer:
                obj2 = convert_branch(obj)
                self.add_branch(obj2)  # call this again, but this time it is not a Branch object
            else:
                print('Omitting branch ' + obj.name)
        else:
            raise Exception('Unrecognized branch type ' + obj.device_type.value)

    def delete_branch(self, obj: Branch):
        """
        Delete a :ref:`Branch<branch>` object from the grid.

        Arguments:

            **obj** (:ref:`Branch<branch>`): :ref:`Branch<branch>` object
        """
        for branch_list in self.get_branch_lists():
            try:
                branch_list.remove(obj)
            except:
                pass

    def delete_line(self, obj: Line):
        """
        Delete line
        :param obj: Line instance
        """
        self.lines.remove(obj)

    def delete_dc_line(self, obj: DcLine):
        """
        Delete line
        :param obj: Line instance
        """
        self.dc_lines.remove(obj)

    def delete_transformer2w(self, obj: Transformer2W):
        """
        Delete transformer
        :param obj: Transformer2W instance
        """
        self.transformers2w.remove(obj)

    def delete_hvdc_line(self, obj: HvdcLine):
        """
        Delete HVDC line
        :param obj:
        """
        self.hvdc_lines.remove(obj)

    def delete_vsc_converter(self, obj: VSC):
        """
        Delete VSC
        :param obj: VSC Instance
        """
        self.vsc_devices.remove(obj)

    def delete_upfc_converter(self, obj: UPFC):
        """
        Delete VSC
        :param obj: VSC Instance
        """
        self.upfc_devices.remove(obj)

    def add_load(self, bus: Bus, api_obj=None):
        """
        Add a :ref:`Load<load>` object to a :ref:`Bus<bus>`.

        Arguments:

            **bus** (:ref:`Bus<bus>`): :ref:`Bus<bus>` object

            **api_obj** (:ref:`Load<load>`): :ref:`Load<load>` object
        """
        if api_obj is None:
            api_obj = Load()
        api_obj.bus = bus

        if self.time_profile is not None:
            api_obj.create_profiles(self.time_profile)

        if api_obj.name == 'Load':
            api_obj.name += '@' + bus.name

        bus.loads.append(api_obj)

        return api_obj

    def add_generator(self, bus: Bus, api_obj=None):
        """
        Add a (controlled) :ref:`Generator<generator>` object to a :ref:`Bus<bus>`.

        Arguments:

            **bus** (:ref:`Bus<bus>`): :ref:`Bus<bus>` object

            **api_obj** (:ref:`Generator<generator>`): :ref:`Generator<generator>`
            object
        """
        if api_obj is None:
            api_obj = Generator()
        api_obj.bus = bus

        if self.time_profile is not None:
            api_obj.create_profiles(self.time_profile)

        bus.controlled_generators.append(api_obj)

        return api_obj

    def add_static_generator(self, bus: Bus, api_obj=None):
        """
        Add a :ref:`StaticGenerator<static_generator>` object to a :ref:`Bus<bus>`.

        Arguments:

            **bus** (:ref:`Bus<bus>`): :ref:`Bus<bus>` object

            **api_obj** (:ref:`StaticGenerator<static_generator>`):
            :ref:`StaticGenerator<static_generator>` object
        """
        if api_obj is None:
            api_obj = StaticGenerator()
        api_obj.bus = bus

        if self.time_profile is not None:
            api_obj.create_profiles(self.time_profile)

        bus.static_generators.append(api_obj)

        return api_obj

    def add_external_grid(self, bus: Bus, api_obj=None):
        """
        Add a :ref:`Load<load>` object to a :ref:`Bus<bus>`.

        Arguments:

            **bus** (:ref:`Bus<bus>`): :ref:`Bus<bus>` object

            **api_obj** (:ref:`Load<load>`): :ref:`Load<load>` object
        """
        if api_obj is None:
            api_obj = ExternalGrid()
        api_obj.bus = bus

        if self.time_profile is not None:
            api_obj.create_profiles(self.time_profile)

        if api_obj.name == 'External grid':
            api_obj.name += '@' + bus.name

        bus.loads.append(api_obj)

        return api_obj

    def add_battery(self, bus: Bus, api_obj=None):
        """
        Add a :ref:`Battery<battery>` object to a :ref:`Bus<bus>`.

        Arguments:

            **bus** (:ref:`Bus<bus>`): :ref:`Bus<bus>` object

            **api_obj** (:ref:`Battery<battery>`): :ref:`Battery<battery>` object
        """
        if api_obj is None:
            api_obj = Battery()
        api_obj.bus = bus

        if self.time_profile is not None:
            api_obj.create_profiles(self.time_profile)

        bus.batteries.append(api_obj)

        return api_obj

    def add_shunt(self, bus: Bus, api_obj=None):
        """
        Add a :ref:`Shunt<shunt>` object to a :ref:`Bus<bus>`.

        Arguments:

            **bus** (:ref:`Bus<bus>`): :ref:`Bus<bus>` object

            **api_obj** (:ref:`Shunt<shunt>`): :ref:`Shunt<shunt>` object
        """
        if api_obj is None:
            api_obj = Shunt()
        api_obj.bus = bus

        if self.time_profile is not None:
            api_obj.create_profiles(self.time_profile)

        bus.shunts.append(api_obj)

        return api_obj

    def add_wire(self, obj: Wire):
        """
        Add Wire to the collection
        :param obj: Wire instance
        """
        if obj is not None:
            if type(obj) == Wire:
                self.wire_types.append(obj)
            else:
                print('The template is not a wire!')

    def delete_wire(self, i, catalogue_to_check=None):
        """
        Delete wire from the collection
        :param i: index
        :param catalogue_to_check: list of catalogue types to check this deletion
        """

        if catalogue_to_check is not None:
            if self.wire_types[i] in catalogue_to_check:
                return False  # do not delete

        self.wire_types.pop(i)
        return True

    def add_overhead_line(self, obj: Tower):
        """
        Add overhead line (tower) template to the collection
        :param obj: Tower instance
        """
        if obj is not None:
            if type(obj) == Tower:
                self.overhead_line_types.append(obj)
            else:
                print('The template is not an overhead line!')

    def delete_overhead_line(self, i, catalogue_to_check=None):
        """
        Delete tower from the collection
        :param i: index
        :param catalogue_to_check: list of catalogue types to check this deletion
        """

        if catalogue_to_check is not None:
            if self.overhead_line_types[i] in catalogue_to_check:
                return False  # do not delete

        self.overhead_line_types.pop(i)
        return True

    def add_underground_line(self, obj: UndergroundLineType):
        """
        Add underground line
        :param obj: UndergroundLineType instance
        """
        if obj is not None:
            if type(obj) == UndergroundLineType:
                self.underground_cable_types.append(obj)
            else:
                print('The template is not an underground line!')

    def delete_underground_line(self, i, catalogue_to_check=None):
        """
        Delete underground line
        :param i: index
        :param catalogue_to_check: list of catalogue types to check this deletion
        """
        if catalogue_to_check is not None:
            if self.underground_cable_types[i] in catalogue_to_check:
                return False  # do not delete

        self.underground_cable_types.pop(i)
        return True

    def add_sequence_line(self, obj: SequenceLineType):
        """
        Add sequence line to the collection
        :param obj: SequenceLineType instance
        """
        if obj is not None:
            if type(obj) == SequenceLineType:
                self.sequence_line_types.append(obj)
            else:
                print('The template is not a sequence line!')

    def delete_sequence_line(self, i, catalogue_to_check=None):
        """
        Delete sequence line from the collection
        :param i: index
        :param catalogue_to_check: list of catalogue types to check this deletion
        """
        if catalogue_to_check is not None:
            if self.sequence_line_types[i] in catalogue_to_check:
                return False  # do not delete

        self.sequence_line_types.pop(i)
        return True

    def add_transformer_type(self, obj: TransformerType):
        """
        Add transformer template
        :param obj: TransformerType instance
        """
        if obj is not None:
            if type(obj) == TransformerType:
                self.transformer_types.append(obj)
            else:
                print('The template is not a transformer!')

    def delete_transformer_type(self, i, catalogue_to_check):
        """
        Delete transformer type from the collection
        :param i: index
        :param catalogue_to_check: list of catalogue types to check this deletion
        """
        if catalogue_to_check is not None:
            if self.transformer_types[i] in catalogue_to_check:
                return False  # do not delete

        self.transformer_types.pop(i)
        return True

    def apply_all_branch_types(self):
        """
        Apply all the branch types
        """
        logger = Logger()
        for branch in self.lines:
            if branch.template is not None:
                branch.apply_template(branch.template, self.Sbase, logger=logger)

        for branch in self.transformers2w:
            if branch.template is not None:
                branch.apply_template(branch.template, self.Sbase, logger=logger)

        return logger

    def add_substation(self, obj: Substation):
        """
        Add substation
        :param obj: Substation object
        """
        self.substations.append(obj)

    def delete_substation(self, i):
        """
        Delete substation
        :param i: index
        """
        self.substations.pop(i)

    def add_area(self, obj: Area):
        """
        Add area
        :param obj: Area object
        """
        self.areas.append(obj)

    def delete_area(self, i):
        """
        Delete area
        :param i: index
        """
        self.areas.pop(i)

    def add_zone(self, obj: Zone):
        """
        Add zone
        :param obj: Zone object
        """
        self.zones.append(obj)

    def delete_zone(self, i):
        """
        Delete zone
        :param i: index
        """
        self.zones.pop(i)

    def add_country(self, obj: Country):
        """
        Add country
        :param obj: Country object
        """
        self.countries.append(obj)

    def delete_country(self, i):
        """
        Delete country
        :param i: index
        """
        self.countries.pop(i)

    def convert_line_to_hvdc(self, line: Line) -> HvdcLine:
        """
        Convert a line to HVDC, this is the GUI way to create HVDC objects
        :param line: Line instance
        :return: HvdcLine
        """
        hvdc = HvdcLine(bus_from=line.bus_from,
                        bus_to=line.bus_to,
                        name='HVDC Line',
                        active=line.active,
                        rate=line.rate,
                        active_prof=line.active_prof,
                        rate_prof=line.rate_prof)

        # add device to the circuit
        self.add_hvdc(hvdc)

        # delete the line from the circuit
        self.delete_line(line)

        return hvdc

    def convert_line_to_transformer(self, line: Line) -> Transformer2W:
        """
        Convert a line to Transformer
        :param line: Line instance
        :return: Transformer2W
        """
        transformer = Transformer2W(bus_from=line.bus_from,
                                    bus_to=line.bus_to,
                                    name='Transformer',
                                    active=line.active,
                                    rate=line.rate,
                                    r=line.R,
                                    x=line.X,
                                    b=line.B,
                                    active_prof=line.active_prof,
                                    rate_prof=line.rate_prof)

        # add device to the circuit
        self.add_transformer2w(transformer)

        # delete the line from the circuit
        self.delete_line(line)

        return transformer

    def convert_line_to_vsc(self, line: Line) -> VSC:
        """
        Convert a line to voltage source converter
        :param line: Line instance
        :return: Nothing
        """
        vsc = VSC(bus_from=line.bus_from,
                  bus_to=line.bus_to,
                  name='VSC',
                  active=line.active,
                  rate=line.rate,
                  r1=line.R,
                  x1=line.X,
                  Beq=line.B,
                  m=1.0,
                  active_prof=line.active_prof,
                  rate_prof=line.rate_prof)

        # add device to the circuit
        self.add_vsc(vsc)

        # delete the line from the circuit
        self.delete_line(line)

        return vsc

    def convert_line_to_upfc(self, line: Line) -> UPFC:
        """
        Convert a line to voltage source converter
        :param line: Line instance
        :return: Nothing
        """
        upfc = UPFC(bus_from=line.bus_from,
                    bus_to=line.bus_to,
                    name='UPFC',
                    active=line.active,
                    rate=line.rate,
                    rl=line.R,
                    xl=line.X,
                    bl=line.B,
                    active_prof=line.active_prof,
                    rate_prof=line.rate_prof)

        # add device to the circuit
        self.add_upfc(upfc)

        # delete the line from the circuit
        self.delete_line(line)

        return upfc

    def plot_graph(self, ax=None):
        """
        Plot the grid.
        :param ax: Matplotlib axis object
        :return:
        """
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        if self.graph is None:
            self.build_graph()

        nx.draw_spring(self.graph, ax=ax)

    def export_pf(self, file_name, power_flow_results):
        """
        Export power flow results to file.

        Arguments:

            **file_name** (str): Excel file name
        """

        if power_flow_results is not None:
            df_bus, df_branch = power_flow_results.export_all()

            df_bus.index = self.bus_names
            df_branch.index = self.branch_names

            writer = pd.ExcelWriter(file_name)
            df_bus.to_excel(writer, 'Bus results')
            df_branch.to_excel(writer, 'Branch results')
            writer.save()
        else:
            raise Exception('There are no power flow results!')

    def export_profiles(self, file_name):
        """
        Export object profiles to file.

        Arguments:

            **file_name** (str): Excel file name
        """

        if self.time_profile is not None:

            # collect data
            P = list()
            Q = list()
            Ir = list()
            Ii = list()
            G = list()
            B = list()
            P_gen = list()
            V_gen = list()
            E_batt = list()

            load_names = list()
            gen_names = list()
            bat_names = list()

            for bus in self.buses:

                for elm in bus.loads:
                    load_names.append(elm.name)
                    P.append(elm.P_prof)
                    Q.append(elm.Q_prof)

                    Ir.append(elm.Ir_prof)
                    Ii.append(elm.Ii_prof)

                    G.append(elm.G_prof)
                    B.append(elm.B_prof)

                for elm in bus.controlled_generators:
                    gen_names.append(elm.name)

                    P_gen.append(elm.P_prof)
                    V_gen.append(elm.Vset_prof)

                for elm in bus.batteries:
                    bat_names.append(elm.name)
                    gen_names.append(elm.name)
                    P_gen.append(elm.P_prof)
                    V_gen.append(elm.Vsetprof)
                    E_batt.append(elm.energy_array)

            # form DataFrames
            P = pd.DataFrame(data=np.array(P).transpose(), index=self.time_profile, columns=load_names)
            Q = pd.DataFrame(data=np.array(Q).transpose(), index=self.time_profile, columns=load_names)
            Ir = pd.DataFrame(data=np.array(Ir).transpose(), index=self.time_profile, columns=load_names)
            Ii = pd.DataFrame(data=np.array(Ii).transpose(), index=self.time_profile, columns=load_names)
            G = pd.DataFrame(data=np.array(G).transpose(), index=self.time_profile, columns=load_names)
            B = pd.DataFrame(data=np.array(B).transpose(), index=self.time_profile, columns=load_names)
            P_gen = pd.DataFrame(data=np.array(P_gen).transpose(), index=self.time_profile, columns=gen_names)
            V_gen = pd.DataFrame(data=np.array(V_gen).transpose(), index=self.time_profile, columns=gen_names)
            E_batt = pd.DataFrame(data=np.array(E_batt).transpose(), index=self.time_profile, columns=bat_names)

            writer = pd.ExcelWriter(file_name)
            P.to_excel(writer, 'P loads')
            Q.to_excel(writer, 'Q loads')

            Ir.to_excel(writer, 'Ir loads')
            Ii.to_excel(writer, 'Ii loads')

            G.to_excel(writer, 'G loads')
            B.to_excel(writer, 'B loads')

            P_gen.to_excel(writer, 'P generators')
            V_gen.to_excel(writer, 'V generators')

            E_batt.to_excel(writer, 'Energy batteries')
            writer.save()
        else:
            raise Exception('There are no time series!')

    def set_state(self, t):
        """
        Set the profiles state at the index t as the default values.
        """
        for bus in self.buses:
            bus.set_state(t)

        for branch in self.get_branches():
            branch.set_profile_values(t)

    def get_bus_branch_connectivity_matrix(self):
        """
        Get the branch-bus connectivity
        :return: Cf, Ct, C
        """
        n = len(self.buses)
        m = self.get_branch_number()
        Cf = lil_matrix((m, n))
        Ct = lil_matrix((m, n))

        bus_dict = {bus: i for i, bus in enumerate(self.buses)}

        for branch_list in self.get_branch_lists():
            for k, br in enumerate(branch_list):
                i = bus_dict[br.bus_from]  # store the row indices
                j = bus_dict[br.bus_to]  # store the row indices
                Cf[k, i] = 1
                Ct[k, j] = 1
        Cf = csc_matrix(Cf)
        Ct = csc_matrix(Ct)
        C = Cf + Ct
        return Cf, Ct, C

    def get_adjacent_matrix(self):
        """
        Get the bus adjacent matrix
        :return: Adjacent matrix
        """
        Cf, Ct, C = self.get_bus_branch_connectivity_matrix()
        A = C.T * C
        return A

    @staticmethod
    def get_adjacent_buses(A: csc_matrix, bus_idx):
        """
        Return array of indices of the buses adjacent to the bus given by it's index
        :param A: Adjacent matrix
        :param bus_idx: bus index
        :return: array of adjacent bus indices
        """
        return A.indices[A.indptr[bus_idx]:A.indptr[bus_idx + 1]]

    def try_to_fix_buses_location(self, buses_selection):
        """
        Try to fix the location of the null-location buses
        :param buses_selection: list of tuples index, bus object
        :return: indices of the corrected buses
        """
        delta = 1e20

        while delta > 10:

            A = self.get_adjacent_matrix()

            for k, bus in buses_selection:

                idx = list(self.get_adjacent_buses(A, k))

                # remove the elements already in the selection
                for i in range(len(idx) - 1, 0, -1):
                    if k == idx[i]:
                        idx.pop(i)

                x_arr = list()
                y_arr = list()
                for i in idx:
                    x_arr.append(self.buses[i].graphic_obj.x())
                    y_arr.append(self.buses[i].graphic_obj.y())

                x_m = np.mean(x_arr)
                y_m = np.mean(y_arr)

                delta_i = np.sqrt((bus.x - x_m)**2 + (bus.y - y_m)**2)

                if delta_i < delta:
                    delta = delta_i

                bus.x = x_m.copy()
                bus.y = y_m.copy()
                bus.graphic_obj.set_position(x=bus.x, y=bus.y)

        print('Done!')

    def get_center_location(self):
        """
        Get the mean coordinates of the system (lat, lon)
        """
        coord = np.array([b.get_coordinates() for b in self.buses])

        return coord.mean(axis=0).tolist()

    def get_boundaries(self):
        """
        Get the graphic representation boundaries
        :return: min_x, max_x, min_y, max_y
        """
        min_x = sys.maxsize
        min_y = sys.maxsize
        max_x = -sys.maxsize
        max_y = -sys.maxsize

        # shrink selection only
        for bus in self.buses:
            bus.retrieve_graphic_position()
            x = bus.x
            y = bus.y
            max_x = max(max_x, x)
            min_x = min(min_x, x)
            max_y = max(max_y, y)
            min_y = min(min_y, y)

        return min_x, max_x, min_y, max_y

    def average_separation(self):
        """
        Average separation of the buses
        :param branches: list of Branch elements
        :return: average separation
        """
        separation = 0.0
        branch_lists = self.get_branch_lists()
        n = 0
        for branch_lst in branch_lists:
            for branch in branch_lst:
                s = np.sqrt((branch.bus_from.x - branch.bus_to.x)**2 + (branch.bus_from.y - branch.bus_to.y)**2)
                separation += s
                n += 1
        return separation / n

    def add_circuit(self, circuit: "MultiCircuit", angle):
        """
        Add a circuit to this circuit
        :param circuit: Circuit to insert
        :param angle: angle in degrees
        :return: Nothing
        """

        min_x, max_x, min_y, max_y = self.get_boundaries()

        sep1 = self.average_separation()

        # compute the average point
        xm = (max_x + min_x) / 2
        ym = (max_y + min_y) / 2

        # compute the radius
        r = np.sqrt((max_x-xm)**2 + (max_y-ym)**2)
        a = np.deg2rad(angle)

        # compute the zero point at which to insert the circuit
        x0 = xm + r * np.cos(a)
        y0 = xm + r * np.sin(a)

        # modify the coordinates of the new circuit
        min_x2, max_x2, min_y2, max_y2 = self.get_boundaries()
        sep2 = self.average_separation()
        factor = sep2 / sep1
        for bus in circuit.buses:
            bus.x = x0 + (bus.x - min_x2) * factor
            bus.y = y0 + (bus.y - min_y2) * factor

        # add profiles if required
        if self.time_profile is not None:

            for bus in circuit.buses:
                bus.create_profiles(index=self.time_profile)

            for lst in [circuit.lines, circuit.transformers2w, circuit.hvdc_lines]:
                for branch in lst:
                    branch.create_profiles(index=self.time_profile)

        self.add_devices_list(self.buses, circuit.buses)
        self.add_devices_list(self.lines, circuit.lines)
        self.add_devices_list(self.transformers2w, circuit.transformers2w)
        self.add_devices_list(self.hvdc_lines, circuit.hvdc_lines)
        self.add_devices_list(self.vsc_devices, circuit.vsc_devices)
        self.add_devices_list(self.dc_lines, circuit.dc_lines)

        return circuit.buses

    def add_devices_list(self, original_list, new_list):
        """
        Add a list of devices to another keeping coherence
        :param original_list:
        :param new_list:
        :return:
        """
        existing_uuids = {e.idtag for e in original_list}

        for elm in new_list:
            if elm.idtag in existing_uuids:
                print(elm.name , 'uuid is repeated..generating new one')
                elm.generate_uuid()
            original_list.append(elm)

    def snapshot_balance(self):
        """
        Creates a report DataFrame with the snapshot active power balance
        :return: DataFrame
        """

        data = {'Generators': 0.0,
                'Static generators': 0.0,
                'Batteries': 0.0,
                'Loads': 0.0,
                'Balance': 0.0}

        for bus in self.buses:

            for gen in bus.controlled_generators:
                if gen.active:
                    data['Generators'] = data['Generators'] + gen.P

            for gen in bus.static_generators:
                if gen.active:
                    data['Static generators'] = data['Static generators'] + gen.P

            for gen in bus.batteries:
                if gen.active:
                    data['Batteries'] = data['Batteries'] + gen.P

            for load in bus.loads:
                if load.active:
                    data['Loads'] = data['Loads'] + load.P

        generation = data['Generators'] + data['Static generators'] + data['Batteries']
        load = data['Loads']
        data['Generation - Load'] = generation - data['Loads']
        data['Imbalance (%)'] = abs(load - generation) / max(load, generation) * 100.0

        return pd.DataFrame(data, index=['Power (MW)']).transpose()

    def scale_power(self, factor):
        """
        Modify the loads and generators
        :param factor: multiplier
        :return: Nothing
        """
        for elm in self.get_loads():
            elm.P *= factor
            elm.Q *= factor

        for elm in self.get_generators():
            elm.P *= factor

        for elm in self.get_static_generators():
            elm.P *= factor
            elm.Q *= factor

    def get_used_templates(self):
        """
        Get a list of the used templates in the objects
        :return: list
        """
        val = set()

        branches = self.get_branches()

        for branch in branches:
            if hasattr(branch, 'template'):
                obj = getattr(branch, 'template')
                val.add(obj)

                # if it is a tower, add the wire templates too
                if obj.device_type == DeviceType.TowerDevice:
                    for wire in obj.wires_in_tower:
                        val.add(wire)

        return list(val)

    def get_automatic_precision(self):
        """
        Get the precision that simulates correctly the power flow
        :return: tolerance parameter for the power flow options, exponent
        """
        injections = list()
        devices = self.get_loads() + self.get_static_generators() + self.get_generators() + self.get_batteries()
        for elm in devices:
            if elm.P != 0.0:
                injections.append(elm.P)

        P = np.abs(injections) / self.Sbase
        P = P[P > 0]
        lg = np.log10(P)
        lg[lg == -np.inf] = 1e20
        exponent = int(np.min(np.abs(lg))) * 3
        tolerance = 1.0 / (10.0 ** exponent)

        return tolerance, exponent

    def fill_xy_from_lat_lon(self, destructive=True, factor=0.01, remove_offset=True):
        """
        fill the x and y value from the latitude and longitude values
        :param destructive: if true, the values are overwritten regardless, otherwise only if x and y are 0
        :param factor: Explosion factor
        :param remove_offset: remove the sometimes huge offset coming from pyproj
        """

        n = len(self.buses)
        lon = np.zeros(n)
        lat = np.zeros(n)
        for i, bus in enumerate(self.buses):
            lon[i] = bus.longitude
            lat[i] = bus.latitude

        # perform the coordinate transformation
        logger = Logger()
        try:
            import pyproj
        except ImportError:
            logger.add_error("pyproj is not installed")
            return logger

        x, y = pyproj.Transformer.from_crs(4326, 25830, always_xy=True).transform(lon, lat)
        x *= factor
        y *= factor

        # remove the offset
        if remove_offset:
            x_min = np.min(x)
            y_max = np.max(y)
            x -= x_min + 100  # 100 is a healthy offset
            y -= y_max - 100  # 100 is a healthy offset

        # assign the values
        for i, bus in enumerate(self.buses):
            if destructive or (bus.x == 0.0 and bus.y == 0.0):
                bus.x = x[i]
                bus.y = -y[i]

        return logger

    def fill_lat_lon_from_xy(self, destructive=True, factor=1.0, offset_x=0, offset_y=0):
        """
        Convert the coordinates to some random lat lon
        :param destructive:
        :param factor:
        :param offset_x:
        :param offset_y:
        :return:
        """
        n = len(self.buses)
        x = np.zeros(n)
        y = np.zeros(n)
        for i, bus in enumerate(self.buses):
            x[i] = bus.x * factor + offset_x
            y[i] = bus.y * factor + offset_y

        logger = Logger()
        try:
            import pyproj
        except ImportError:
            logger.add_error("pyproj is not installed")
            return logger

        proj_latlon = pyproj.Proj(proj='latlong', datum='WGS84')
        proj_xy = pyproj.Proj(proj="utm", zone=33, datum='WGS84')
        lonlat = pyproj.transform(proj_xy, proj_latlon, x, y)

        lon = lonlat[0]
        lat = lonlat[1]

        # assign the values
        for i, bus in enumerate(self.buses):
            if destructive or (bus.x == 0.0 and bus.y == 0.0):
                bus.latitude = lat[i]
                bus.longitude = lon[i]

        return logger

    def import_bus_lat_lon(self, df: pd.DataFrame, bus_col, lat_col, lon_col):
        """

        :param df: Pandas DataFrame with the information
        :param bus_col: bus column name
        :param lat_col: latitude column name
        :param lon_col: longitude column name
        :return:
        """
        logger = Logger()
        lats = df[lat_col].values
        lons = df[lon_col].values
        names = df[bus_col].values

        d = dict()
        for lat, lon, name in zip(lats, lons, names):
            d[str(name)] = (lat, lon)

        # assign the values
        for i, bus in enumerate(self.buses):
            if bus.name in d.keys():
                lat, lon = d[bus.name]
                bus.latitude = lat
                bus.longitude = lon
            elif bus.code in d.keys():
                lat, lon = d[bus.code]
                bus.latitude = lat
                bus.longitude = lon
            else:
                logger.add_error("No coordinates for bus", bus.name)

        return logger

    def import_plexos_load_profiles(self, df: pd.DataFrame):
        """

        :param df:
        :return: Logger
        """
        logger = Logger()
        nn = df.shape[0]
        if self.get_time_number() != nn:
            self.format_profiles(df.index.values)

        df.columns = [val.split('_')[0] for val in df.columns.values]

        bus_by_code = {bus.code: bus for bus in self.buses}

        for col_name in df.columns.values:
            try:
                bus = bus_by_code[col_name]
                for i, load in enumerate(bus.loads):
                    if i == 0:
                        load.P_prof = df[col_name].values
                        load.Q_prof = load.P_prof * 0.8
                    else:
                        load.P_prof = np.zeros(nn)
                        load.Q_prof = np.zeros(nn)
            except KeyError:
                logger.add_error("Missing in the model", col_name)

        return logger

    def import_plexos_generation_profiles(self, df: pd.DataFrame):
        """

        :param df:
        :return: Logger
        """
        logger = Logger()
        nn = df.shape[0]
        if self.get_time_number() != nn:
            self.format_profiles(df.index.values)

        df.columns = [val.replace('GEN_', '') for val in df.columns.values]

        generators = self.get_generators()
        gen_by_name = {gen.name: gen for gen in generators}

        for col_name in df.columns.values:
            try:
                gen = gen_by_name[col_name]
                gen.P_prof = df[col_name].values
            except KeyError:
                logger.add_error("Missing in the model", col_name)

        return logger

    def import_branch_rates_profiles(self, df: pd.DataFrame):
        """

        :param df:
        :return: Logger
        """
        logger = Logger()
        nn = df.shape[0]
        if self.get_time_number() != nn:
            self.format_profiles(df.index.values)

        # substitute the stupid psse names by their equally stupid short names
        # 11000_AGUAYO_400_12004_ABANTO_400_1_CKT
        cols = list()
        for val in df.columns.values:
            vals = val.split('_')
            if len(vals) < 7:
                logger.add_error("Wrong PSSe name", val)
                cols.append(val)
            else:
                col = vals[0] + '_' + vals[3] + '_' + vals[6]
                cols.append(col)
        df.columns = cols

        branches = self.get_branches()
        elm_by_name = {elm.name: elm for elm in branches}

        for col_name in df.columns.values:
            try:
                elm = elm_by_name[col_name]
                elm.rate_prof = df[col_name].values
            except KeyError:
                # log the error but keep the default rate
                logger.add_error("Missing in the model", col_name)

        return logger

    def get_bus_area_indices(self):
        """
        Get array of area indices for each bus
        :return:
        """
        d = {elm: k for k, elm in enumerate(self.areas)}

        lst = np.zeros(len(self.buses), dtype=int)
        for k, bus in enumerate(self.buses):
            lst[k] = d[bus.area]
        return lst

    def get_area_buses(self, area: Area) -> List[Tuple[int, Bus]]:
        """
        Get the selected buses
        :return:
        """
        lst: List[Tuple[int, Bus]] = list()
        for k, bus in enumerate(self.buses):
            if bus.area == area:
                lst.append((k, bus))
        return lst

    def get_areas_buses(self, areas: List[Area]) -> List[Tuple[int, Bus]]:
        """
        Get the selected buses
        :return:
        """
        lst: List[Tuple[int, Bus]] = list()
        for k, bus in enumerate(self.buses):
            if bus.area in areas:
                lst.append((k, bus))
        return lst

    def get_zone_buses(self, zone: Zone) -> List[Tuple[int, Bus]]:
        """
        Get the selected buses
        :return:
        """
        lst: List[Tuple[int, Bus]] = list()
        for k, bus in enumerate(self.buses):
            if bus.zone == zone:
                lst.append((k, bus))
        return lst

    def get_inter_area_branches(self, a1: Area, a2: Area):
        """
        Get the inter-area branches
        :param a1: Area from
        :param a2: Area to
        :return: List of (branch index, branch object, flow sense w.r.t the area exchange)
        """
        lst: List[Tuple[int, object, float]] = list()
        for k, branch in enumerate(self.get_branches()):
            if branch.bus_from.area == a1 and branch.bus_to.area == a2:
                lst.append((k, branch, 1.0))
            elif branch.bus_from.area == a2 and branch.bus_to.area == a1:
                lst.append((k, branch, -1.0))
        return lst

    def get_inter_areas_branches(self, a1: List[Area], a2: List[Area]):
        """
        Get the inter-area branches. HVDC branches are not considered
        :param a1: Area from
        :param a2: Area to
        :return: List of (branch index, branch object, flow sense w.r.t the area exchange)
        """
        lst: List[Tuple[int, object, float]] = list()
        for k, branch in enumerate(self.get_branches_wo_hvdc()):
            if branch.bus_from.area in a1 and branch.bus_to.area in a2:
                lst.append((k, branch, 1.0))
            elif branch.bus_from.area in a2 and branch.bus_to.area in a1:
                lst.append((k, branch, -1.0))
        return lst

    def get_inter_areas_hvdc_branches(self, a1: List[Area], a2: List[Area]):
        """
        Get the inter-area branches
        :param a1: Area from
        :param a2: Area to
        :return: List of (branch index, branch object, flow sense w.r.t the area exchange)
        """
        lst: List[Tuple[int, object, float]] = list()
        for k, branch in enumerate(self.hvdc_lines):
            if branch.bus_from.area in a1 and branch.bus_to.area in a2:
                lst.append((k, branch, 1.0))
            elif branch.bus_from.area in a2 and branch.bus_to.area in a1:
                lst.append((k, branch, -1.0))
        return lst

    def get_inter_zone_branches(self, z1: Zone, z2: Zone):
        """
        Get the inter-area branches
        :param z1: Zone from
        :param z2: Zone to
        :return: List of (branch index, branch object, flow sense w.r.t the area exchange)
        """
        lst: List[Tuple[int, object, float]] = list()
        for k, branch in enumerate(self.get_branches()):
            if branch.bus_from.zone == z1 and branch.bus_to.zone == z2:
                lst.append((k, branch, 1.0))
            elif branch.bus_from.zone == z2 and branch.bus_to.zone == z1:
                lst.append((k, branch, -1.0))
        return lst

    def get_branch_area_connectivity_matrix(self, a1: List[Area], a2: List[Area]):
        """
        Get the inter area connectivity matrix
        :param a1: list of sending areas
        :param a2: list of receiving areas
        :return: Connectivity of the branches to each sending or receiving area groups (branches, 2)
        """
        area_dict = {a: i for i, a in enumerate(self.areas)}

        area1_list = [area_dict[a] for a in a1]
        area2_list = [area_dict[a] for a in a2]

        branches = self.get_branches()  # all including HVDC

        conn = lil_matrix((len(branches), 2), dtype=int)

        for k, elm in enumerate(branches):
            i = area_dict[elm.bus_from.area]
            j = area_dict[elm.bus_to.area]
            if i != j:
                if (i in area1_list) and (j in area2_list):
                    # from->to matches the areas
                    conn[k, 0] = 1
                    conn[k, 1] = -1
                elif (i in area2_list) and (j in area1_list):
                    # reverse the sign
                    conn[k, 0] = -1
                    conn[k, 1] = 1

        return conn.tocsc()

    def change_base(self, Sbase_new):
        """
        Change the elements base impedance
        :param Sbase_new: new base impedance in MVA
        """
        Sbase_old = self.Sbase

        # get all the branches with impedance
        elms = self.get_branches_wo_hvdc()

        # change the base at each element
        for elm in elms:
            elm.change_base(Sbase_old, Sbase_new)

        # assign the new base
        self.Sbase = Sbase_new

    def fuse_devices(self):
        """
        Fuse all the different devices in a node to a single device per node
        :return:
        """
        for bus in self.buses:
            bus.fuse_devices()

    def re_index_time(self, year=None, hours_per_step=1):
        """
        Generate sequential time steps to correct the time_profile
        :param year: base year, if None, this year is taken
        :param hours_per_step: number of hours per step, by default 1 hour by step
        """
        if year is None:
            t0 = datetime.now()
            year = t0.year

        nt = self.get_time_number()
        t0 = datetime(year=year, month=1, day=1)
        tm = [t0 + timedelta(hours=t * hours_per_step) for t in range(nt)]
        self.time_profile = pd.to_datetime(tm)

    def set_generators_active_profile_from_their_active_power(self):
        """
        Modify the generators active profile to match the active power profile
        if P=0, active = False else active=True
        """
        for g in self.get_generators():
            g.active_prof = g.P_prof.astype(bool)

    def set_batteries_active_profile_from_their_active_power(self):
        """
        Modify the batteries active profile to match the active power profile
        if P=0, active = False else active=True
        """
        for g in self.get_batteries():
            g.active_prof = g.P_prof.astype(bool)

    def correct_inconsistencies(self, logger: Logger, maximum_difference=0.1, min_vset=0.98, max_vset=1.02):
        """
        Correct devices' inconsistencies
        :param logger: logger to store the events
        :param maximum_difference: proportion to be under or above (i.e.
                                   Transformer HV=41.9, bus HV=45 41.9/45 = 0.93 ->
                                   0.9 <= 0.93 <= 1.1, so its ok
        :param min_vset: minimum voltage set point (p.u.)
        :param max_vset: maximum voltage set point (p.u.)
        :return:
        """
        for elm in self.transformers2w:
            elm.fix_inconsistencies(logger,
                                    maximum_difference=maximum_difference)

        for elm in self.get_generators():
            elm.fix_inconsistencies(logger,
                                    min_vset=min_vset,
                                    max_vset=max_vset)

    def normalize_bus_positions_offset(self, base_offset=100):
        """
        Normalize the massive offset that may be in the Qt graphic objects' positions
        :param base_offset:
        :return:
        """
        n = len(self.buses)
        x = np.zeros(n, dtype=int)
        y = np.zeros(n, dtype=int)

        # read values
        for i, bus in enumerate(self.buses):
            gr = bus.graphic_obj
            x[i] = gr.x()
            y[i] = gr.y()

        # correct values
        x_min = np.min(x)
        y_max = np.max(y)
        x -= x_min + base_offset  # 100 is a healthy offset
        y -= y_max - base_offset  # 100 is a healthy offset

        # assign values
        for i, bus in enumerate(self.buses):
            bus.x = x[i]
            bus.y = y[i]
            if bus.graphic_obj is not None:
                bus.graphic_obj.set_position(x[i], y[i])
